import matplotlib.pyplot as plt
import numpy as np

from .. import fastq
from .. import utilities
from .. import adapters
from . import optional_ax, igv_colors, blues

@optional_ax
def plot_quality_histograms(quality_counts, ax=None):
    ''' Plots full distribution of quality values seens across all reads
        at each cycle.
    '''
    quality_counts = np.array(quality_counts, dtype=np.float)

    image = ax.imshow(quality_counts.T,
                      origin='lower',
                      interpolation='nearest',
                      cmap=blues,
                     )
    ax.set_xlabel('Cycle index')
    ax.set_ylabel('Quality score')

    return ax.figure

def plot_paired_quality_histograms(stats):
    num_cycles, num_q_scores = stats['R1_qs'].shape
    fig, (R1_ax, R2_ax) = plt.subplots(2, 1, figsize=(0.1 * num_cycles, 0.1 * (num_q_scores * 2 + 10)))
    plot_quality_histograms(stats['R1_qs'], ax=R1_ax)
    plot_quality_histograms(stats['R2_qs'], ax=R2_ax)
    R1_ax.set_title('R1')
    R2_ax.set_title('R2')

def plot_joint_average_quality_distribution(joint_average_q_distribution):
    fig, ax = plt.subplots(figsize=(12, 8))
    ax.imshow(joint_average_q_distribution,
              origin='lower',
              interpolation='nearest',
              cmap=blues,
             )
    ax.set_xlabel('R2 average quality')
    ax.set_ylabel('R1 average quality')
    
def plot_data_statistics(stats,
                         R1_expected_seq=None,
                         R2_expected_seq=None,
                         R1_bracket_ranges=None,
                         R2_bracket_ranges=None,
                        ):
    ''' Plot fractions of all base calls that are each base at each cycle.
    '''
    fig, (R1_ax, R2_ax) = plt.subplots(2, 1, figsize=(20, 10))

    plot_composition(stats['R1_cs'], ax=R1_ax, expected_seq=R1_expected_seq, bracket_ranges=R1_bracket_ranges)
    plot_composition(stats['R2_cs'], ax=R2_ax, expected_seq=R2_expected_seq, bracket_ranges=R2_bracket_ranges)
    R1_ax.set_title('R1')
    R2_ax.set_title('R2')

    for ax in (R1_ax, R2_ax):
        ax.set_ylabel('Base composition')

    R1_q_ax = R1_ax.twinx()
    R2_q_ax = R2_ax.twinx()
    
    R1_mean_qs = [utilities.mean_from_histogram(row) for row in stats['R1_qs']]
    R2_mean_qs = [utilities.mean_from_histogram(row) for row in stats['R2_qs']]

    style = {'linestyle': '-',
             'color': 'black',
             'linewidth': 2,
             'alpha': 0.3,
            }

    R1_q_ax.plot(R1_mean_qs, **style)
    R2_q_ax.plot(R2_mean_qs, **style)

    for ax in [R1_q_ax, R2_q_ax]:
        ax.set_ylabel('Average quality score', rotation=90 + 180, va='bottom')
        ax.set_ylim(0, 41)
    
    total = stats['R1_cs'].sum(axis=1)[0]
    plt.suptitle('Base composition vs. cycle index\n{0:,d} total reads'.format(total))

@optional_ax
def plot_composition(base_counts,
                     bases_before=0,
                     expected_seq=None,
                     bracket_ranges=None,
                     ax=None,
                     save_as=None,
                     match_threshold=0.95,
                     as_percentage=False,
                     bases_to_plot=utilities.base_order,
                    ):
    ''' Plot fractions of all base calls that are each base at each cycle.
    '''
    cycles, _ = base_counts.shape
    denominators = np.maximum(1, base_counts.sum(axis=1)).astype(float)

    base_order = utilities.base_order
    
    xs = np.arange(-bases_before, cycles - bases_before)
    
    for i, base in enumerate(base_order):
        if base not in bases_to_plot:
            continue

        fractions = base_counts[:, i] / denominators
        if as_percentage:
            fractions *= 100

        line_style = {'linestyle': '-',
                      'linewidth': 0.5,
                      'alpha': 0.5,
                      'color': igv_colors[base],
                     }
        marker_style = {'marker': '.',
                        'color': igv_colors[base],
                        'label': base,
                        'linestyle': 'None',
                       }

        ax.plot(xs, fractions, **marker_style) 

        if not expected_seq:
            ax.plot(xs, fractions, **line_style) 
    
    if expected_seq:
        start, sequence = expected_seq
        shade_background(start, sequence, observed_seq=base_counts, match_threshold=match_threshold, ax=ax)

    if bracket_ranges:
        for text, (start, end) in bracket_ranges:
            # Truncate brackets that would extend off of the axis.
            if start > max(xs):
                continue
            end = min(end, max(xs) + 1)
            draw_range_bracket(ax, start, end - 1, text)
    
    y_max = 1.01
    y_line = 1

    if as_percentage:
        y_max *= 100
        y_line *= 100

    ax.set_ylim(0, y_max)
    ax.axhline(y=y_line, color='black', alpha=0.5)
    ax.set_xlim(min(xs) - 0.5, max(xs) + 0.5)

    ax.legend(loc='center left',
              bbox_to_anchor=(1.04, 0.5),
              framealpha=0.5,
              numpoints=1,
             )

@optional_ax
def shade_background(start, sequence, observed_seq=None, match_threshold=0.95, ax=None, save_as=None):
    ''' Lightly shade the background according to the expected sequence.
    '''
    for p, expected_bases in enumerate(sequence):
        expected_bases = [k for k, _ in utilities.group_by(expected_bases)]

        alpha = 0.2

        if observed_seq is not None:
            if len(expected_bases) == 1:
                expected_base = expected_bases[0]
                observed_frac = observed_seq[p][utilities.base_to_index[expected_base]] / max(observed_seq[p].sum(), 1)
                if observed_frac > match_threshold:
                    alpha = 0.1
                else:
                    alpha = 0.4
        else:
            alpha = 0.1

        increment = 1. / len(expected_bases)
        for i, base in enumerate(expected_bases):
            ax.axvspan(start + p - 0.5,
                       start + p + 0.5,
                       ymax=1 - i * increment,
                       ymin=1 - (i + 1) * increment,
                       facecolor=igv_colors[base],
                       alpha=alpha,
                       linewidth=0.7,
                      )


@optional_ax
def plot_adapter_composition(base_counts,
                             bases_before,
                             expected_seq,
                             adapter_ranges,
                             qualities,
                             ax=None,
                             save_as=None,
                            ):
    num_after_start = min(120, len(qualities) - bases_before)
    plot_composition(base_counts[:bases_before + num_after_start],
                     bases_before=bases_before,
                     ax=ax,
                     expected_seq=(0, expected_seq),
                     bracket_ranges=adapter_ranges,
                    )
    
    q_ax = ax.twinx()

    mean_qs = [utilities.mean_from_histogram(row) for row in qualities[:bases_before + num_after_start]]

    style = {'linestyle': '-',
             'color': 'black',
             'linewidth': 2,
             'alpha': 0.3,
            }
    q_ax.set_autoscale_on(False)
    xs = np.arange(-bases_before, num_after_start)
    q_ax.plot(xs, mean_qs, **style)
    q_ax.set_ylim(0, 41)

def plot_mean_position_qs(position_type_counts, xs, ax):
    position_average_qs = variants.compute_average_qualities(position_type_counts)
    _, max_qual_index, _, _ = position_type_counts.shape

    style = {'linestyle': '-',
             'color': 'black',
             'linewidth': 1,
             'alpha': 0.3,
            }
    
    ax.set_autoscale_on(False)
    ax.plot(xs, position_average_qs, **style)
    ax.set_ylim(0, max_qual_index - 1)
    ax.set_ylabel('Average quality score')
    
@optional_ax
def plot_average_qualities_by_basecall(average_q_distribution, label=None, save_as=None, ax=None, expected_seq=None):
    average_q_distribution = np.array(average_q_distribution, dtype=float)
    #average_q_distribution /= average_q_distribution.sum()

    cycles, _ = average_q_distribution.shape
    
    xs = np.arange(cycles)

    for i, base in enumerate(utilities.base_order):
        fractions = average_q_distribution[:, i]
        line_style = {'linestyle': '-',
                      'linewidth': 0.5,
                      'alpha': 0.5,
                      'color': igv_colors[base],
                     }
        marker_style = {'marker': '.',
                        'color': igv_colors[base],
                        'label': base,
                        'linestyle': 'None',
                       }
        ax.plot(xs, fractions, **marker_style) 
        if not expected_seq:
            ax.plot(xs, fractions, **line_style) 

    if expected_seq:
        start, sequence = expected_seq
        shade_background(start, sequence, ax=ax)

    ax.set_xlim(min(xs) - 0.5, max(xs) + 0.5)
    ax.set_xlabel('Cycle')
    ax.set_ylabel('Average q score')

    ax.legend(loc='center left',
              bbox_to_anchor=(1.04, 0.5),
              framealpha=0.5,
              numpoints=1,
             )

@optional_ax
def plot_average_qualities(qs, label=None, save_as=None, ax=None):
    mean_from_histogram = np.vectorize(utilities.mean_from_histogram, signature='(m)->()')
    means = mean_from_histogram(qs)
    ax.plot(means, '-', color='grey', label=label)
    ax.set_xlim(0, len(means) - 1)
    ax.set_xlabel('Cycle')
    ax.set_ylabel('Average q score')

@optional_ax
def plot_paired_average_qualities(stats, name=None, save_as=None, ax=None):
    if name == None:
        R1_label = 'R1'
        R2_label = 'R2'
    else:
        R1_label = '{}_R1'.format(name)
        R2_label = '{}_R2'.format(name)
    plot_average_qualities(stats['R1_qs'], label=R1_label, ax=ax)
    plot_average_qualities(stats['R2_qs'], label=R2_label, ax=ax)
    ax.legend(loc='upper left', framealpha=0.5)

def remaining_Ns(base_counts, fig_file_name, paired=False):
    ''' Plot fractions of all base calls that are each base at each cycle.
    '''
    cycles, _ = base_counts.shape
    denominators = np.maximum(1, base_counts.sum(axis=1)).astype(float)
    total = int(denominators[0])

    fig = plt.figure()
    if paired:
        R1_counts = base_counts[:cycles / 2, :]
        R2_counts = base_counts[cycles / 2:, :]
        axs = [fig.add_subplot(2, 1, 1), fig.add_subplot(2, 1, 2)]
    else:
        raise NotImplementedError

    bases = fastq.base_order
    colors = 'mgbcrr'
    for ax, counts in zip(axs, [R1_counts, R2_counts]):
        counts_remaining = counts[::-1].cumsum(axis=0)[::-1]
        denoms = counts_remaining.sum(axis=1)
        fractions_remaining = np.true_divide(counts_remaining, denoms[:, np.newaxis])
        for i, (base, color) in enumerate(zip(bases, colors)):
            ax.plot(fractions_remaining[:, i], 'o-', color=color, linewidth=0.5, label=base) 
        ax.set_ylim(0, 1)
        ax.legend(loc='upper right', framealpha=0.5)
    
    plt.suptitle('Base composition vs. cycle index\n{0:,d} total reads'.format(total))
    fig.set_size_inches( (15, 15) )
    #plt.savefig(fig_file_name)
    #plt.close()



import argparse
import subprocess
import time
import os
import shutil
import re
import logging
import pickle
import glob
from collections import defaultdict
from pathlib import Path

import yaml

from . import split_file
from . import split_sra_file
from . import launcher
from .. import serialize, fastq, utilities

def extend_stages(whole_stages, specific_stages):
    for whole_stage, specific_stage in zip(whole_stages, specific_stages):
        whole_stage.extend(specific_stage)

class MapReduceExperiment(object):
    specific_results_files = [
        ('summary', serialize.log),
        ('log', '', 'log.txt'),
    ]
    specific_figure_files = []
    specific_outputs = []
    specific_work = []
    specific_cleanup = []

    def __init__(self, **kwargs):
        self.group = kwargs.get('group', '')
        self.name = kwargs['name']
        
        self.data_dir = Path(kwargs.get('data_dir', '/dev/null'))

        home = Path.home()

        self.work_prefix = Path(kwargs.get('work_prefix', home))
        self.scratch_prefix = Path(kwargs.get('scratch_prefix', home / 'scratch'))
        if 'results_dir' in kwargs:
            self.relative_results_dir = Path(kwargs['results_dir']).relative_to(self.work_prefix)
        elif 'relative_results_dir' in kwargs:
            self.relative_results_dir = Path(kwargs['relative_results_dir'])
        else:
            raise ValueError('must provide either results_dir or relative_results_dir')

        self.num_pieces = kwargs['num_pieces']
        self.which_piece = kwargs['which_piece']
        
        suffix = split_file.generate_suffix(self.num_pieces, self.which_piece)

        self.scratch_results_dir = (self.scratch_prefix / self.relative_results_dir).with_suffix(suffix)
        self.work_results_dir = self.work_prefix / self.relative_results_dir
        
        if self.which_piece == -1:
            # Sentinel value that indicates this is the merged experiment.
            # Only create the directory in this instance to avoid race
            # conditions.
            if not os.path.isdir(self.work_results_dir):
                os.makedirs(self.work_results_dir)
        else:
            # Only need to create the scratch directory if this ISN't the merged
            # experiment.
            if not os.path.isdir(self.scratch_results_dir):
                os.makedirs(self.scratch_results_dir)

        # Build file name templates and lists of outputs, work, and cleanup from
        # the specific lists of these things specified by each member in the
        # inheritance chain.
        class_hierarchy = [c for c in self.__class__.__mro__[::-1] if c != object]
        self.results_files = []
        self.figure_files = []
        self.outputs = [[] for _ in range(self.num_stages)]
        self.work = [[] for _ in range(self.num_stages)]
        self.cleanup = [[] for _ in range(self.num_stages)]
        for c in class_hierarchy:
            self.results_files.extend(c.specific_results_files)
            self.figure_files.extend(c.specific_figure_files)
            extend_stages(self.outputs, c.specific_outputs)
            extend_stages(self.work, c.specific_work)
            extend_stages(self.cleanup, c.specific_cleanup)
    
        for stage in range(self.num_stages):
            for kind in ['summary_stage', 'timing']:
                key = '{0}_{1}'.format(kind, stage)
                tail_template = '{{name}}_{0}_{1}.txt'.format(kind, stage)
                self.results_files.append((key, serialize.log, tail_template))
                self.outputs[stage].append(key)

        self.cleanup[-1].append('consolidate_summaries')

        self.make_file_names()
        self.summary = []

        if self.which_piece == -1:
            piece_string = ' '
        else:
            piece_string = ' {0:{length}} / {1} '.format(self.which_piece, self.num_pieces, length=len(str(self.num_pieces)))

        format_string = '%(asctime)s{0}%(message)s'.format(piece_string)

        logging.basicConfig(filename=self.file_names['log'],
                            level=logging.INFO,
                            format=format_string, 
                           )

    @classmethod
    def from_description_file_name(cls, description_file_name,
                                   num_pieces=1,
                                   which_piece=-1,
                                   **extra_kwargs):
        description = parse_description(description_file_name)
        description.update(extra_kwargs)
        return cls(num_pieces=num_pieces, which_piece=which_piece, **description)

    def make_file_names(self):
        self.file_names = {}
        self.merged_file_names = {}
        self.file_types = {}

        for file_info in self.results_files:
            if len(file_info) == 3:
                key, serialize_type, tail_template = file_info
            else:
                key, serialize_type = file_info
                tail_template = '{name}_{key}.{extension}'

            if isinstance(serialize_type, tuple):
                serialize_type, fast_merge = serialize_type
            else:
                fast_merge = False

            if isinstance(serialize_type, str):
                extension = serialize_type
            else:
                extension = serialize_type.extension

            file_tail = tail_template.format(name=self.name, key=key, extension=extension)
            self.file_names[key] = self.scratch_results_dir / file_tail
            self.merged_file_names[key] = self.work_results_dir / file_tail
            self.file_types[key] = (serialize_type, fast_merge)

        self.figure_file_names = {}
        for key, tail_template in self.figure_files:
            file_tail = tail_template.format(name=self.name, key=key)
            self.figure_file_names[key] = self.work_results_dir / file_tail

        if self.which_piece == -1:
            self.file_names = self.merged_file_names

    def format_file_name(self, key, **kwargs):
        return Path(str(self.file_names[key]).format(**kwargs))

    def consolidate_summaries(self):
        stage_file_names = [self.file_names['summary_stage_{0}'.format(stage)]
                            for stage in range(self.num_stages)]
        serialize.log.consolidate_stages(stage_file_names, self.file_names['summary'])

    def write_file(self, key, data):
        file_format, _ = self.file_types[key]
        file_name = self.file_names[key]

        if file_format == 'pickle':
            with open(file_name, 'wb') as file_handle:
                pickle.dump(data, file_handle)
        else:
            file_format.write_file(data, file_name)

    def read_file(self, key, merged=False, **kwargs):
        if merged == False:
            file_name = self.file_names[key]
        else:
            file_name = self.merged_file_names[key]
        
        file_format, _ = self.file_types[key]

        if file_format == 'pickle':
            data = pickle.load(open(file_name, 'rb'))
        elif file_format == 'txt':
            data = open(file_name).read()
        else:
            data = file_format.read_file(file_name, **kwargs)

        return data

    def do_work(self, stage):
        logging.info('Beginning work for stage {0}'.format(stage))

        times = []
        for function_name in self.work[stage]:
            logging.info('Starting function {0}'.format(function_name))
            start_time = time.time()
            self.__getattribute__(function_name)()
            end_time = time.time()
            times.append((function_name, end_time - start_time))

        self.write_file('timing_{0}'.format(stage), times)
        self.write_file('summary_stage_{0}'.format(stage), self.summary)
        
        logging.info('Done with work for stage {0}'.format(stage))

    def do_cleanup(self, stage):
        logging.info('Beginning cleanup for stage {0}'.format(stage))

        times = []
        for function_name in self.cleanup[stage]:
            logging.info('Starting function {0}'.format(function_name))
            start_time = time.time()
            self.__getattribute__(function_name)()
            end_time = time.time()
            times.append((function_name, end_time - start_time))
        
        serialize.log.append(times, self.merged_file_names['timing_{0}'.format(stage)])
        logging.info('Done with cleanup for stage {0}'.format(stage))

    def get_fn_groups(self):
        fn_groups = defaultdict(lambda: {'R1': None,
                                         'R2': None,
                                         'I1': None,
                                         'I2': None,
                                        })
        pattern = r'[RI][12]'

        data_fns = list(self.data_dir.glob('*.fastq')) + list(self.data_dir.glob('*.fq'))
        for data_fn in data_fns:
            which_member = re.search(pattern, data_fn.name).group()
            group_name = re.sub(pattern, '*', data_fn.name)
            fn_groups[group_name][which_member] = data_fn
        
        for group_name in sorted(fn_groups):
            R1_fn = fn_groups[group_name]['R1']
            R2_fn = fn_groups[group_name]['R2']

            if R1_fn == None or R2_fn == None:
                raise ValueError('unpaired file names in data_dir')

        return fn_groups

    @utilities.memoized_property
    def data_fns(self):
        data_fns = []
        for extension in ['fastq', 'fq', 'sra', 'fastq.gz']:
            data_fns.extend(glob.glob(f'{self.data_dir}/*.{extension}'))
        return data_fns
    
    def get_reads(self, max_reads_per_file=None, quiet=False):
        ''' A generator over the reads in a piece of each data file.
            Can handle a mixture of different fastq encodings across (but not
            within) files.
        '''
        total_reads = 0
        for file_name in self.data_fns:
            total_reads_from_file = 0

            if file_name.endswith('sra'):
                file_piece = split_sra_file.piece(file_name,
                                                  self.num_pieces,
                                                  self.which_piece,
                                                 )
            else:
                if self.which_piece == -1:
                    file_piece = file_name
                else:
                    file_piece = split_file.piece(file_name,
                                                  self.num_pieces,
                                                  self.which_piece,
                                                  'fastq',
                                                 )

            reads = fastq.reads(file_piece,
                                standardize_names=self.standardize_names,
                                ensure_sanger_encoding=self.standardize_names,
                               )
            for read in reads:
                yield read
                
                total_reads += 1
                total_reads_from_file += 1

                if not quiet and total_reads % 100000 == 0:
                    logging.info('{0:,} reads processed'.format(total_reads))

                if max_reads_per_file and total_reads_from_file >= max_reads_per_file:
                    break
            
            head, tail = os.path.split(file_name)
            if not quiet:
                self.summary.append(('Reads in {0}'.format(tail), total_reads_from_file))
            
        if not quiet:
            logging.info('{0:,} total reads processed'.format(total_reads))
            self.summary.append(('Total reads', total_reads))
    
    def get_read_pairs(self, max_pairs_per_group=None, quiet=False):
        fn_groups = self.get_fn_groups()

        read_pairs_list = []

        for group in sorted(fn_groups):
            R1_fn = fn_groups[group]['R1']
            R2_fn = fn_groups[group]['R2']

            if R1_fn == None or R2_fn == None:
                raise ValueError('unpaired file names in data_dir')

            R1_lines = split_file.interleaved_piece(R1_fn, self.num_pieces, self.which_piece)
            R2_lines = split_file.interleaved_piece(R2_fn, self.num_pieces, self.which_piece)
            read_pairs = fastq.read_pairs(R1_lines, R2_lines,
                                          standardize_names=True,
                                          ensure_sanger_encoding=True,
                                         )
            read_pairs_list.append(read_pairs)
        
        if len(fn_groups) == 0:
            sra_fns = glob.glob(self.data_dir + '/*.sra')
            for sra_fn in sra_fns:
                lines = split_sra_file.piece(sra_fn,
                                             self.num_pieces,
                                             self.which_piece,
                                             paired=True,
                                            )
                read_pairs = fastq.read_pairs_interleaved(lines,
                                                          standardize_names=True,
                                                          ensure_sanger_encoding=True,
                                                         )
                read_pairs_list.append(read_pairs)
        
        total_reads = 0

        for read_pairs in read_pairs_list:
            total_pairs_from_group = 0
            for R1, R2 in read_pairs:
                if R1.name != R2.name:
                    raise ValueError('R1 and R2 out of sync', R1, R2)

                yield R1, R2
                
                total_reads += 1
                total_pairs_from_group += 1

                if not quiet and total_reads % 100000 == 0:
                    logging.info('{0:,} reads processed'.format(total_reads))

                if max_pairs_per_group and total_pairs_from_group >= max_pairs_per_group:
                    break
        
        if not quiet:
            logging.info('{0:,} total reads processed'.format(total_reads))
            self.summary.append(('Total reads', total_reads))


def controller(ExperimentClass, script_path, **override):
    args = parse_arguments()
    if args.subparser_name == 'launch':
        if override:
            print('Launching with {0}'.format(override))
        launch(args, script_path, ExperimentClass.num_stages, **override)
    elif args.subparser_name == 'process':
        process(args, ExperimentClass, **override)
    elif args.subparser_name == 'finish':
        finish(args, ExperimentClass, **override)

def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument('--job_dir',
                        required=True,
                       )

    subparsers = parser.add_subparsers(dest='subparser_name')

    parser_process = subparsers.add_parser('process')
    parser_process.add_argument('--num_pieces', type=int,
                                help='how many pieces',
                                default=1,
                               )
    parser_process.add_argument('--which_piece', type=int,
                                help='which piece this is',
                                default=0,
                               )
    parser_process.add_argument('--stage', type=int,
                                help='stage',
                                default=0,
                               )
    
    parser_launch = subparsers.add_parser('launch')
    parser_launch.add_argument('--num_pieces', type=int,
                               default=1,
                              )

    parser_finish = subparsers.add_parser('finish')
    parser_finish.add_argument('--num_pieces', type=int,
                                help='how many pieces',
                                default=1,
                               )
    parser_finish.add_argument('--stage', type=int,
                                help='stage',
                                default=0,
                               )

    args = parser.parse_args()
    return args

def get_description_fn(job_dir):
    description_fn = Path(job_dir) / 'description.yaml'
    if not description_fn.exists():
        description_fn = Path(job_dir) / 'description.txt'

    if not description_fn.exists():
        raise ValueError(job_dir)

    return description_fn

def parse_description(description_fn, **override):
    description_fn = Path(description_fn)
    if description_fn.suffix == '.txt':
        description = dict(line.strip().split() for line in open(description_fn)
                           if not line.startswith('#'))
    elif description_fn.suffix == '.yaml':
        description = yaml.load(description_fn.read_text())

    description.update(override)

    return description

def launch(args, script_path, num_stages, **override):
    description_fn = Path(args.job_dir) / 'description.yaml'
    if not description_fn.exists():
        description_fn = Path(args.job_dir) / 'description.txt'

    if not description_fn.exists():
        raise ValueError

    description = parse_description(description_fn)

    def make_process_command(args, which_piece, stage):
        command = [script_path,
                   '--job_dir', args.job_dir,
                   'process',
                   '--num_pieces', str(args.num_pieces),
                   '--which_piece', str(which_piece),
                   '--stage', str(stage),
                  ]
        command_string = ' '.join(command) + '\n'
        return command_string
    
    def make_finish_command(args, stage):
        command = [script_path,
                   '--job_dir', args.job_dir,
                   'finish',
                   '--num_pieces', str(args.num_pieces),
                   '--stage', str(stage),
                  ]
        command_string = ' '.join(command) + '\n'
        return command_string

    process_file_names = []
    finish_file_names = []
    job_names = []

    for stage in range(num_stages):
        job_name = '{0}_{1}_{2}'.format(description['name'],
                                        args.num_pieces,
                                        stage,
                                       )
        job_names.append(job_name)

        process_file_name = '{0}/process_{1}_stage_{2}'.format(args.job_dir,
                                                               args.num_pieces,
                                                               stage,
                                                              )
        process_file_names.append(process_file_name)

        finish_file_name = '{0}/finish_{1}_stage_{2}'.format(args.job_dir,
                                                             args.num_pieces,
                                                             stage,
                                                            )
        finish_file_names.append(finish_file_name)
                              
        with open(process_file_name, 'w') as process_file:
            for which_piece in range(args.num_pieces):
                line = make_process_command(args, which_piece, stage)
                process_file.write(line)

        with open(finish_file_name, 'w') as finish_file:
            line = make_finish_command(args, stage)
            finish_file.write(line)

    # Launch with qsub on lonestar, sbatch on stampede, parallel otherwise.
    hostname = os.environ.get('HOSTNAME', '')
    if 'tacc' in hostname:
        if 'ls4' in hostname:
            submitter = 'qsub'
            def get_job_id(output):
                return re.search(r'Your job (\d+)', output).group(1)
        elif 'stampede' in hostname:
            submitter = 'sbatch'
            def get_job_id(output):
                return re.search(r'Submitted batch job (\d+)', output).group(1)

        starting_path = os.getcwd()
        os.chdir(args.job_dir)
        launcher_file_name = launcher.create(
            job_names[0],
            process_file_names[0],
            time='00:20:00',
            optional_finish='bash {0}'.format(finish_file_names[0]),
        )
        output = subprocess.check_output([submitter, launcher_file_name])
        this_job_id = get_job_id(output)
        print('\tLaunched stage {0} with jid {1}'.format(stage, this_job_id))

        for stage in range(1, num_stages):
            previous_job_id = this_job_id
            launcher_file_name = launcher.create(
                job_names[stage],
                process_file_names[stage],
                time='00:20:00',
                optional_finish='bash {0}'.format(finish_file_names[stage]),
                hold_jid=previous_job_id,
            )
            output = subprocess.check_output([submitter, launcher_file_name])
            this_job_id = get_job_id(output)
            print('\tLaunched stage {0} with jid {1}, holding on {2}'.format(stage,
                                                                           this_job_id,
                                                                           previous_job_id,
                                                                          )
                 )
        os.chdir(starting_path)
    else:
        for stage in range(num_stages):
        #for stage in [1]:
            print('\tLaunched stage {0} with parallel'.format(stage))
            subprocess.check_call('parallel < {0}'.format(process_file_names[stage]), shell=True)
            subprocess.check_call('bash {0}'.format(finish_file_names[stage]), shell=True)

def process(args, ExperimentClass, **override):
    description_fn = get_description_fn(args.job_dir)
    description = parse_description(description_fn, **override)

    experiment = ExperimentClass(num_pieces=args.num_pieces,
                                 which_piece=args.which_piece,
                                 **description)
    experiment.do_work(args.stage)

def finish(args, ExperimentClass, **override):
    description_fn = get_description_fn(args.job_dir)
    description = parse_description(description_fn, **override)
    
    merged = ExperimentClass(num_pieces=args.num_pieces,
                             which_piece=-1,
                             **description)
    pieces = [ExperimentClass(num_pieces=args.num_pieces,
                              which_piece=which_piece,
                              **description)
              for which_piece in range(args.num_pieces)]

    merge_times = []
    for key in merged.outputs[args.stage]:
        piece_file_names = [piece.file_names[key] for piece in pieces]
        merged_file_name = merged.merged_file_names[key]
        file_type, fast_merge = merged.file_types[key]

        logging.info('Merging file {0} (fast_merge={1})'.format(key, fast_merge))
        start_time = time.time()
        serialize.merge_files(piece_file_names,
                              merged_file_name,
                              file_type,
                              fast=fast_merge,
                             )
        end_time = time.time()
        merge_times.append(('Merging {}'.format(key), end_time - start_time))

    serialize.log.append(merge_times, merged.merged_file_names['timing_{0}'.format(args.stage)])
    merged.do_cleanup(args.stage)



import numpy as np

extension = 'npy'

def write_file(array, file_name):
    ''' Writes a 3D array into a text file.
    '''
    np.save(file_name, array)

def read_file(file_name):
    ''' Reads the format written by array_file.write into an array.
    '''
    array = np.load(file_name)
    return array

def combine_data(first_array, second_array):
    ''' Identically shaped arrays are added.
    '''
    if first_array.shape != second_array.shape:
        raise ValueError
    
    combined = first_array + second_array

    return combined



import itertools
import collections

extension = 'txt'

def _write_pairs(pairs, log_file):
    ''' Writes pairs to a scalar log file. Each line has the form
        description:value
        where value is a comma-separated list of ints or floats.
    '''
    for description, value in pairs:
        if not isinstance(value, collections.Iterable):
            value = [value]
        if isinstance(value[0], int):
            to_string = str
        else:
            to_string = '{:.2f}'.format
        value_string = ','.join(map(to_string, value))
        log_file.write('{0}: {1}\n'.format(description, value_string))

def write_file(pairs, log_file_name):
    with open(log_file_name, 'w') as log_file:
        _write_pairs(pairs, log_file)

def append(pairs, log_file_name):
    with open(log_file_name, 'a') as log_file:
        _write_pairs(pairs, log_file)

def read_file(log_file_name):
    ''' Reads the format written by write_file into a list of tuples.
    '''
    def process_line(line):
        description, value = line.strip().rsplit(':', 1)
        strings = value.split(',')
        try:
            int(strings[0])
            dtype = int
        except ValueError:
            dtype = float
        value = [dtype(s) for s in strings]
        return description, value
    
    with open(log_file_name) as log_file:
        pairs = [process_line(l) for l in log_file]

    return pairs

def combine_data(first_pairs, second_pairs):
    combined_pairs = []
    zipped = itertools.zip_longest(first_pairs, second_pairs, fillvalue=(None, None))
    for (first_d, first_v), (second_d, second_v) in zipped:
        if first_d != second_d:
            raise RuntimeError('descriptions do not match')
        
        # Note: the use of '+' means these can't be np.arrays.
        combined_v = first_v + second_v 
        combined_pairs.append((first_d, combined_v))
    
    return combined_pairs

def collapse_pairs(pairs):
    collapsed_pairs = [(d, sum(v)) for d, v in pairs]
    return collapsed_pairs

def consolidate_stages(file_names, consolidated_file_name):
    pairs_list = [read_file(file_name) for file_name in file_names]
    collapsed_pairs_list = [collapse_pairs(pairs) for pairs in pairs_list]
    consolidated_pairs = itertools.chain.from_iterable(collapsed_pairs_list)
    write_file(consolidated_pairs, consolidated_file_name)



from hits import utilities, variants

extension = 'txt'

def _consolidate_counts(positions_list):
    positions_list = sorted(positions_list)
    consolidated_list = []
    for position, items in utilities.group_by(positions_list,
                                              key=lambda x: x[:4],
                                             ):
        ref_seq_name, ref_pos, ref_char, read_char = position
        count = sum(item[4] for item in items)
        consolidated = (ref_seq_name, ref_pos, ref_char, read_char, count)
        consolidated_list.append(consolidated)
    return consolidated_list

def write_file(positions_list, file_name):
    positions_list = _consolidate_counts(positions_list)
    with open(file_name, 'w') as fh:
        for ref_seq_name, ref_pos, ref_char, read_char, count in positions_list:
            line = '{0}\t{1}\t{2}\t{3}\t{4}\n'.format(ref_seq_name,
                                                      ref_pos,
                                                      ref_char,
                                                      read_char,
                                                      count,
                                                     )
            fh.write(line)

def read_file(file_name):
    def line_to_position(line):
        ref_seq_name, ref_pos, ref_char, read_char, count = line.strip().split('\t')
        ref_pos = int(ref_pos)
        count = int(count)
        return ref_seq_name, ref_pos, ref_char, read_char, count

    with open(file_name) as fh:
        positions_list = map(line_to_position, fh)

    return positions_list

def combine_data(first_positions_list, second_positions_list):
    combined_positions_list = first_positions_list + second_positions_list
    return combined_positions_list

def group_by_type(ref_positions_file_name, by_type_file_name):
    # Since write_file does _consolidate_counts, anything produced by read_file
    # will be consolidated.
    positions_list = read_file(ref_positions_file_name)

    type_lists = {change: [] for change in variants.change_order[:12]}

    for ref_seq_name, ref_pos, ref_char, read_char, count in positions_list:
        type_lists[ref_char, read_char].append((ref_seq_name, ref_pos, count))

    with open(by_type_file_name, 'w') as by_type_file:
        for ref_char, read_char in variants.change_order[:12]:
            by_type_file.write('{0} -> {1}\n'.format(ref_char, read_char))
            for ref_seq_name, ref_pos, count in type_lists[ref_char, read_char]:
                line = '\t{0}\t{1}\t{2}\n'.format(ref_seq_name,
                                                  ref_pos,
                                                  count,
                                                 )
                by_type_file.write(line)



def Annotation_factory(fields, read_only=False, **kwargs):
    if read_only:
        curlies = ['{' + name + ':' + spec[-1] + '}' for name, spec in fields]
    else:
        curlies = ['{' + name + ':' + spec.format(**kwargs) + '}'
                   for name, spec in fields
                  ]
    template = '_'.join(curlies).format
    names = [name for name, spec in fields]
    cast_function = {'s': str,
                     'd': int,
                     'f': float,
                    }
    casts = [cast_function[spec[-1]] for name, spec in fields]

    class Annotation(dict):
        @classmethod
        def from_identifier(cls, identifier):
            # Extracts an annotation from a string whose right-most underscore
            # separated fields are the values.
            values = identifier.rsplit('_', len(names) - 1)
            return cls({n: c(v) for n, c, v in zip(names, casts, values)})
        
        @classmethod
        def from_prefix_identifier(cls, identifier):
            # Extracts an annotation from a string whose left-most underscore
            # separated fields are the values.
            values = identifier.split('_', len(names))
            return cls({n: c(v) for n, c, v in zip(names, casts, values)})

        @classmethod
        def from_annotation(cls, annotation):
            return cls({name: annotation[name] for name in names})
        
        @property
        def identifier(self):
            return template(**self)

        def __str__(self):
            return template(**self)

    return Annotation

def make_convertor(encoded_as, convert_to):
    ''' Returns a function that takes a SAM line or AlignedSegment, interprets
        the QNAME field as an identifier for an encoded_as Annotation, and
        converts this to a convert_to Annotation.
    '''
    def converter(line):
        if isinstance(line, str): 
            qname = line.split('\t')[0]
        else:
            qname = line.qname
        encoded = encoded_as.from_identifier(qname)
        converted = convert_to.from_annotation(encoded)
        return converted

    return converter



import array
import sys
import copy
from collections import Counter, defaultdict

import numpy as np
import pysam

from . import utilities
from . import fastq
from . import fasta
from . import annotation
from . import sam
from . import interval
from .sw_cython import *

empty_alignment = {
    'score': -1e6,
    'path': [],
    'query_mappings': [],
    'target_mappings': [],
    'insertions': set(),
    'deletions': set(),
    'mismatches': set(),
}

def first_query_index(alignment_path):
    for q, t in alignment_path:
        if q != GAP:
            return q

    return None

def first_target_index(alignment_path):
    for q, t in alignment_path:
        if t != GAP:
            return t

    return None

def last_query_index(alignment_path):
    for q, t in alignment_path[::-1]:
        if q != GAP:
            return q

    return None

def last_target_index(alignment_path):
    for q, t in alignment_path[::-1]:
        if t != GAP:
            return t

    return None

def last_n_query_pairs(alignment_path, n):
    ''' Returns (q, t) pairs for the last n elements of alignment_path for which
    q is not a gap.
    '''
    pairs = []

    for q, t in alignment_path[::-1]:
        if q != GAP:
            pairs.append((q, t))
        if len(pairs) == n:
            break
    
    pairs = pairs[::-1]

    return pairs

def make_char_pairs(index_pairs, query, target):
    char_pairs = []
    for q, t in index_pairs:
        if q == GAP:
            q_char = '-'
        else:
            q_char = query[q]
        if t == GAP:
            t_char = '-'
        else:
            t_char = target[t]
        char_pairs.append((q_char, t_char))
    return char_pairs

def print_local_alignment(alignment, fh=sys.stdout):
    query = alignment['query']
    target = alignment['target']
    path = alignment['path']
    if path == []:
        fh.write(f'{query}\n\n{target}\n')
        return

    def index_to_char(seq, index):
        if index == GAP:
            return '-'
        else:
            if index >= len(seq):
                raise ValueError(index, len(seq), seq)
            return seq[index]
    
    first_q = first_query_index(path)
    first_t = first_target_index(path)
    last_q = last_query_index(path)
    last_t = last_target_index(path)

    left_query = query[:first_q]
    left_target = target[:first_t]
    right_query = query[last_q + 1:]
    right_target = target[last_t + 1:]

    left_length = max(first_q, first_t)
    left_query = '{:>{length}s} '.format(left_query, length=left_length)
    left_target = '{:>{length}s} '.format(left_target, length=left_length)
    
    query_string = [left_query]
    target_string = [left_target]

    for q, t in path:
        query_string.append(index_to_char(query, q))
        target_string.append(index_to_char(target, t))

    query_string.extend([' ', right_query])
    target_string.extend([' ', right_target])

    fh.write('query\t{0}\n'.format(''.join(query_string)))
    fh.write('target\t{0}\n'.format(''.join(target_string)))

trimmed_annotation_fields = [
    ('original_name', 's'),
    ('insert_length', '04d'),
    ('adapter_seq', 's'),
    ('adapter_qual', 's'),
]

TrimmedAnnotation = annotation.Annotation_factory(trimmed_annotation_fields)
NO_DETECTED_OVERLAP = -2

def generate_alignments(query,
                        target,
                        alignment_type='local',
                        match_bonus=2,
                        mismatch_penalty=-1,
                        indel_penalty=-5,
                        max_alignments=1,
                        min_score=-np.inf,
                        N_matches=True,
                        insertion_penalty=None,
                        deletion_penalty=None,
                       ):

    force_query_start = False
    force_target_start = False
    force_either_start = False

    force_query_end = False
    force_target_end = False
    force_edge_end = False

    if alignment_type == 'local':
        # no forces need to be set to True
        pass

    elif alignment_type == 'edge':
        force_edge_end = True

    elif alignment_type == 'barcode':
        force_query_start = True
        force_target_start = True
        force_edge_end = True

    elif alignment_type == 'overlap':
        force_either_start = True
        force_edge_end = True
    
    elif alignment_type == 'query_end':
        force_query_end = True

    elif alignment_type == 'query_start':
        force_query_start = True

    elif alignment_type == 'unpaired_adapter':
        force_query_start = True
        force_edge_end = True

    elif alignment_type == 'IVT':
        force_query_start = True

    elif alignment_type == 'global':
        force_query_start = True
        force_target_start = True
        force_edge_end = True

    elif alignment_type == 'whole_query':
        force_query_start = True
        force_query_end = True

    elif alignment_type == 'fixed_start':
        force_query_start = True
        force_target_start = True

    elif alignment_type == 'fixed_end':
        force_query_end = True
        force_target_end = True

    query_bytes = query.encode()
    target_bytes = target.encode()

    if indel_penalty is not None:
        insertion_penalty = indel_penalty
        deletion_penalty = indel_penalty

    matrices = generate_matrices(query_bytes,
                                 target_bytes,
                                 match_bonus,
                                 mismatch_penalty,
                                 insertion_penalty,
                                 deletion_penalty,
                                 force_query_start,
                                 force_target_start,
                                 force_either_start,
                                 N_matches,
                                )
    
    cells_seen = set()

    if force_query_end and force_target_end:
        possible_ends = propose_bottom_right_corner(matrices['scores'])
    elif force_edge_end:
        possible_ends = propose_edge_ends(matrices['scores'], cells_seen, min_score, max_alignments=max_alignments)
    elif force_query_end:
        possible_ends = propose_query_edge_ends(matrices['scores'], cells_seen, min_score, max_alignments=max_alignments)
    else:
        possible_ends = propose_all_ends(matrices['scores'], cells_seen, min_score)

    alignments = []
    for end_row, end_col in possible_ends:
        alignment = backtrack_cython(query_bytes,
                                     target_bytes,
                                     matrices,
                                     cells_seen,
                                     end_row,
                                     end_col,
                                     force_query_start,
                                     force_target_start,
                                     force_either_start,
                                    )
        if alignment != None and len(alignment['path']) != 0:
            alignment['query'] = query
            alignment['target'] = target
            alignment['score_ratio'] = alignment['score'] / (len(alignment['path']) * match_bonus)
            alignments.append(alignment)
            if len(alignments) == max_alignments:
                break
        else:
            pass

    return alignments

def propose_bottom_right_corner(score_matrix):
    num_rows, num_cols = score_matrix.shape
    yield (num_rows - 1, num_cols - 1)

def propose_edge_ends(score_matrix,
                      cells_seen,
                      min_score=None,
                      max_alignments=1,
                     ):
    num_rows, num_cols = score_matrix.shape
    if max_alignments == 1:
        max_row = np.argmax(score_matrix[:, num_cols - 1])
        max_row_score = score_matrix[max_row, num_cols -1]
        max_col = np.argmax(score_matrix[num_rows - 1, :])
        max_col_score = score_matrix[num_rows - 1, max_col]
        if max_row_score > max_col_score:
            sorted_edge_cells = [(max_row, num_cols - 1)]
        else:
            sorted_edge_cells = [(num_rows - 1, max_col)]
    else:
        right_edge = [(i, num_cols - 1) for i in range(num_rows)]
        # Note: range(num_cols - 1) prevents including the corner twice
        bottom_edge = [(num_rows - 1, i) for i in range(num_cols - 1)]
        edge_cells = right_edge + bottom_edge
        sorted_edge_cells = sorted(edge_cells,
                                   key=lambda cell: score_matrix[cell],
                                   reverse=True,
                                  )
    for cell in sorted_edge_cells:
        if min_score != None and score_matrix[cell] < min_score:
            break

        if cell in cells_seen:
            continue

        yield cell

def propose_query_edge_ends(score_matrix,
                            cells_seen,
                            min_score=None,
                            max_alignments=1,
                           ):
    num_rows, num_cols = score_matrix.shape
    if max_alignments == 1:
        max_col = np.argmax(score_matrix[num_rows - 1, :])
        sorted_edge_cells = [(num_rows - 1, max_col)]
    else:
        bottom_edge = [(num_rows - 1, i) for i in range(num_cols)]
        sorted_edge_cells = sorted(bottom_edge,
                                   key=lambda cell: score_matrix[cell],
                                   reverse=True,
                                  )
    for cell in sorted_edge_cells:
        if min_score != None and score_matrix[cell] < min_score:
            break

        if cell in cells_seen:
            continue

        yield cell

def propose_all_ends(score_matrix, cells_seen, min_score):
    sorted_indices = score_matrix.ravel().argsort()[::-1]
    for index in sorted_indices:
        cell = np.unravel_index(index, score_matrix.shape)

        if score_matrix[cell] < min_score:
            break

        if cell in cells_seen:
            continue

        yield cell

def global_alignment(query, target, **kwargs):
    al, = generate_alignments(query, target, 'global', **kwargs)
    return al

def infer_insert_length(R1, R2, before_R1, before_R2, indel_penalty=-5, solid=False, in_R1_prefix=None, in_R2_prefix=None):
    ''' Infer the length of the insert represented by R1 and R2 by performing
        a semi-local alignment of R1 and the reverse complement of R2 with
        the expected adapter sequences prepended to each read.
    '''
    # Attempt a fast approach - check if expected sequencing primers occur at the same offset
    # in both reads and the implied insert in each read is close to identical.
    fast_insert_length = None

    if in_R1_prefix is not None and in_R2_prefix is not None:
        if in_R1_prefix in R1.seq and in_R2_prefix in R2.seq:
            R1_end = R1.seq.index(in_R1_prefix)
            R2_end = R2.seq.index(in_R2_prefix)
            if R1_end == R2_end:
                insert_length = R1_end
                R1_trimmed = R1.seq[:insert_length].encode()
                R2_trimmed = utilities.reverse_complement(R2.seq[:insert_length]).encode()
                
                if hamming_distance(R1_trimmed, R2_trimmed) / max(insert_length, 1) < 0.05:
                    fast_insert_length = insert_length

    if fast_insert_length is not None:
        results = {'length': fast_insert_length}
        return results

    # If not, fall back on dynamic programming.

    extended_R1 = before_R1 + R1.seq
    extended_R2 = utilities.reverse_complement(before_R2 + R2.seq)
    alignments = generate_alignments(extended_R1,
                                     extended_R2, 
                                     'overlap',
                                     2,
                                     -1,
                                     indel_penalty,
                                     1,
                                     0,
                                    )

    if len(alignments) == 0:
        results = {'failed': 'no alignments detected'}
        return results
    else:
        alignment = alignments[0]

    results = {'alignment': alignment}

    R1_start = len(before_R1)
    R2_start = len(R2.seq) - 1
    R1_start_in_R2 = alignment['query_mappings'][R1_start]
    R2_start_in_R1 = alignment['target_mappings'][R2_start]
    R1_end_in_R2 = alignment['query_mappings'][len(extended_R1) - 1]
    R2_end_in_R1 = alignment['target_mappings'][0]
    
    # Since R1 is the query and R2 is the target, bases in R1 that aren't in
    # R2 are called insertions, and bases in R2 that aren't in R1 are called
    # deletions.
    # An indel in the insert is non-physical.
    if R2_start_in_R1 != SOFT_CLIPPED:
        illegal_insertion = any(R1_start <= i <= R2_start_in_R1 for i in alignment['insertions'])
    else:
        illegal_insertion = any(R1_start <= i for i in alignment['insertions'])

    if R1_start_in_R2 != SOFT_CLIPPED:
        illegal_deletion = any(R1_start_in_R2 <= d <= R2_start for d in alignment['deletions'])
    else:
        illegal_deletion = any(d <= R2_start for d in alignment['deletions'])
    
    if illegal_insertion or illegal_deletion:
        if illegal_insertion:
            results['failed'] = 'illegal insertion'
        elif illegal_deletion:
            results['failed'] = 'illegal insertion'

        return results

    if len(alignment['path']) == 0:
        results['failed'] = 'no alignment'
        return results

    if R1_start_in_R2 != SOFT_CLIPPED and R2_start_in_R1 != SOFT_CLIPPED:
        length_from_R1 = R2_start_in_R1 - R1_start + 1
        length_from_R2 = R2_start - R1_start_in_R2 + 1
    elif R1_start_in_R2 != SOFT_CLIPPED and R1_end_in_R2 != SOFT_CLIPPED:
        # R1 entirely contained in R2
        length_from_R2 = len(R2) - R1_start_in_R2
        length_from_R1 = length_from_R2
    elif R2_start_in_R1 != SOFT_CLIPPED and R2_end_in_R1 != SOFT_CLIPPED:
        # R2 entirely contained in R1
        length_from_R1 = R2_start_in_R1 - R1_start + 1
        length_from_R2 = length_from_R1
    else:
        # overlap alignment forces the alignment to start with either the
        # beginning of R1 or R2 and end with either the end of R1 or R2. 
        # Making it to this else branch means that either the first base of R1 or
        # the first base of R2 or both wasn't aligned. This either means that
        # the insert is longer than the read length or a pathological alignment
        # has been produced in which only adapter bases are involved in the 
        # alignment. Flag the second case as illegal.

        try:
            first_R1_index, first_R2_index = alignment['path'][0]
        except IndexError:
            print(R1)
            print(R2)
            print(alignment)
            raise
        length_from_R1 = (first_R1_index - R1_start + 1) + (len(R2.seq) - 1)

        last_R1_index, last_R2_index = alignment['path'][-1]
        length_from_R2 = (R2_start - last_R2_index + 1) + (len(R1.seq) - 1)
        
        if first_R1_index == 0 or last_R2_index == 0:
            results['failed'] = 'only adapters'
            return results

    if length_from_R1 < -1 or length_from_R2 < -1:
        # Negative insert lengths are non-physical. Even though I don't
        # understand it, -1 is relatively common so is tolerated.
        results['failed'] = 'negative insert length'
        return results

    if alignment['score_ratio'] < 0.7:
        results['failed'] = 'bad score'
    else:
        if length_from_R1 != length_from_R2:
            if solid and not(alignment['insertions'] or alignment['deletions']):
                pass
            else:
                # This shouldn't be possible without an illegal indel.
                results['failed'] = 'length mismatch'
        else:
            results['length'] = length_from_R1
    
    return results

def print_diagnostic(R1, R2, before_R1, before_R2, alignment, fh=sys.stdout):
    extended_R1 = (before_R1.lower() + R1.seq).decode()
    extended_R2 = (utilities.reverse_complement(before_R2.lower() + R2.seq)).decode()
    #fh.write(R1.name + '\n')
    #fh.write(R1.qual + '\n')
    #fh.write(R2.qual + '\n')
    #fh.write('{0}\t{1}\t{2}\n'.format(alignment['score'], len(alignment['path']) * .2, alignment['score'] - len(alignment['path']) * 2))
    #fh.write(str(alignment['path']) + '\n')
    print_local_alignment(extended_R1, extended_R2, alignment['path'], fh=fh)
    #fh.write(str(alignment['insertions']) + '\n')
    #fh.write(str(alignment['deletions']) + '\n')
    #fh.write(str(sorted(alignment['mismatches'])) + '\n')
    #for q, t in sorted(alignment['mismatches']):
    #    fh.write('\t{0}\t{1}\n'.format(extended_R1[q], extended_R2[t]))

def align_read(read, targets, min_path_length, header,
               min_score_ratio=0.8,
               max_alignments_per_target=1,
               both_directions=True,
               read_interval=None,
               ref_intervals=None,
               **kwargs):

    if read_interval is None:
        read_interval = interval.Interval(0, len(read) - 1)

    if ref_intervals is None:
        ref_intervals = {}

    for target_name, target_seq in targets:
        if target_name not in ref_intervals:
            ref_intervals[target_name] = interval.Interval(0, len(target_seq) - 1)

    skipped_at_start = read_interval.start
    skipped_at_end = len(read) - 1 - read_interval.end

    # Register full seq and quals before possibly trimming read
    original_forward_seq = read.seq
    original_forward_qual = array.array('B', fastq.decode_sanger(read.qual))

    rc_read = read.reverse_complement()
    original_rc_seq = rc_read.seq
    original_rc_qual = array.array('B', fastq.decode_sanger(rc_read.qual))

    # Trim read to read interval
    read = read[read_interval.start:read_interval.end + 1]

    forward_seq = read.seq

    rc_read = read.reverse_complement()
    rc_seq = rc_read.seq

    alignments = []

    if both_directions:
        directions = [False, True]
    else:
        directions = [False]

    for target_name, target_seq in targets:
        ref_interval = ref_intervals[target_name]
        target_seq = target_seq[ref_interval.start:ref_interval.end + 1]

        target_alignments = []

        for is_reverse in directions:
            if is_reverse:
                seq = rc_seq

                original_seq = original_rc_seq
                original_qual = original_rc_qual

                extra_soft_clip_start = skipped_at_end
                extra_soft_clip_end = skipped_at_start

            else:
                seq = forward_seq

                original_seq = original_forward_seq
                original_qual = original_forward_qual

                extra_soft_clip_start = skipped_at_start
                extra_soft_clip_end = skipped_at_end

            for alignment in generate_alignments(seq, target_seq, max_alignments=max_alignments_per_target, **kwargs):
                path = alignment['path']

                if len(path) >= min_path_length and alignment['score_ratio'] >= min_score_ratio:
                    al = pysam.AlignedSegment(header)
                    al.seq = original_seq
                    al.query_qualities = original_qual
                    al.is_reverse = is_reverse

                    char_pairs = make_char_pairs(path, seq, target_seq)

                    cigar = sam.aligned_pairs_to_cigar(char_pairs)

                    clip_from_start = first_query_index(path) + extra_soft_clip_start
                    if clip_from_start > 0:
                        cigar = [(sam.BAM_CSOFT_CLIP, clip_from_start)] + cigar

                    clip_from_end = len(seq) - 1 - last_query_index(path) + extra_soft_clip_end
                    if clip_from_end > 0:
                        cigar = cigar + [(sam.BAM_CSOFT_CLIP, clip_from_end)]

                    al.cigar = cigar
                    
                    if al.query_length != al.infer_query_length():
                        raise ValueError(f'CIGAR implies different query length - {al.query_name}: {al.query_length}, {al.infer_query_length()}')

                    # NOTE: some evidence this might mess up if alignment ends in an indel.
                    md = sam.aligned_pairs_to_MD_string(char_pairs)
                    al.set_tag('MD', md)

                    al.set_tag('AS', alignment['score'])
                    al.reference_name = target_name
                    al.query_name = read.name
                    al.next_reference_id = -1

                    # Add back target seq that was ignored
                    al.reference_start = first_target_index(path) + ref_interval.start

                    target_alignments.append(al)

        target_alignments = sorted(target_alignments, key=lambda al: al.get_tag('AS'), reverse=True)
        alignments.extend(target_alignments[:max_alignments_per_target])

    return alignments

def align_reads(target_fasta_fn,
                reads,
                bam_fn,
                min_path_length=15,
                error_fn='/dev/null',
                alignment_type='overlap',
                yield_unaligned=False,
               ):
    ''' Aligns reads to targets in target_fasta_fn by Smith-Waterman, storing
    alignments in bam_fn and yielding unaligned reads.
    '''
    header = sam.header_from_fasta(target_fasta_fn)

    targets = fasta.to_dict(target_fasta_fn)
    targets = {r: targets[r] for r in header.references}

    alignment_sorter = sam.AlignmentSorter(bam_fn, header)

    statistics = Counter()
    
    def generator():
        with alignment_sorter:
            for read in reads:
                statistics['input'] += 1

                alignments = align_read(read, targets, alignment_type, min_path_length, header)
                
                if alignments:
                    statistics['aligned'] += 1

                    sorted_alignments = sorted(alignments, key=lambda m: m.get_tag('AS'), reverse=True)
                    grouped = utilities.group_by(sorted_alignments, key=lambda m: m.get_tag('AS'))
                    _, highest_group = next(grouped)
                    primary_already_assigned = False
                    for alignment in highest_group:
                        if len(highest_group) == 1:
                            alignment.mapping_quality = 2
                        else:
                            alignment.mapping_quality = 1

                        if not primary_already_assigned:
                            primary_already_assigned = True
                        else:
                            alignment.is_secondary = True

                        alignment_sorter.write(alignment)
                else:
                    statistics['unaligned'] += 1

                    yield read

            with open(error_fn, 'w') as error_fh:
                for key in ['input', 'aligned', 'unaligned']:
                    error_fh.write(f'{key}: {statistics[key]:,}\n')

    if yield_unaligned:
        return generator()
    else:
        for _ in generator():
            pass

def stitch_read_pair(R1, R2, before_R1='', before_R2='', indel_penalty=-5):
    if len(before_R1) >= 10 and len(before_R2) >= 10:
        in_R1_prefix = utilities.reverse_complement(before_R2)[:10]
        in_R2_prefix = utilities.reverse_complement(before_R1)[:10]
    else:
        in_R1_prefix = None
        in_R2_prefix = None

    results = infer_insert_length(R1, R2, before_R1, before_R2, indel_penalty, in_R1_prefix=in_R1_prefix, in_R2_prefix=in_R2_prefix)

    if 'failed' in results:
        insert_length = len(R1) + len(R2)
    else:
        insert_length = results['length']

    R2_rc = R2.reverse_complement()

    R1_overlap_start = max(0, insert_length - len(R2))
    just_R1 = R1[:R1_overlap_start]
    overlap_R1 = R1[R1_overlap_start:insert_length]

    R2_overlap_start = max(0, len(R2) - insert_length)
    overlap_R2 = R2_rc[R2_overlap_start:R2_overlap_start + len(overlap_R1)]
    just_R2 = R2_rc[R2_overlap_start + len(overlap_R1):]

    overlap_seq = []
    overlap_qual = []
    for R1_s, R1_q, R2_s, R2_q in zip(overlap_R1.seq,
                                      overlap_R1.qual,
                                      overlap_R2.seq,
                                      overlap_R2.qual,
                                     ):
        if R1_q > R2_q:
            s, q = R1_s, R1_q
        else:
            s, q = R2_s, R2_q

        overlap_seq.append(s)
        overlap_qual.append(q)

    overlap = fastq.Read('', ''.join(overlap_seq), ''.join(overlap_qual))

    stitched = just_R1 + overlap + just_R2

    return stitched

class SeedAndExtender():
    def __init__(self, target_seq_bytes, max_seed_length, header, target_name):
        if not isinstance(target_seq_bytes, bytes):
            target_seq_bytes = target_seq_bytes.encode()
            
        self.target_seq_bytes = target_seq_bytes
        self.max_seed_length = max_seed_length
        self.header = header
        self.target_name = target_name
        
        seed_locations = defaultdict(set)
        
        for k in range(1, max_seed_length + 1):
            for start in range(len(target_seq_bytes) - k + 1):
                k_mer = target_seq_bytes[start:start + k]
                seed_locations[k_mer].add(start)
                
        self.seed_locations = seed_locations

    def seed_and_extend(self, query_seq_bytes, true_query_start, true_query_end, query_name):
        ''' true_query_end should point one past the end of the seed '''
        if not isinstance(query_seq_bytes, bytes):
            query_seq_bytes = query_seq_bytes.encode()

        # If the requested seed is longer than max_seed_length, truncate it, then check if
        # the extended truncated seed at least covers the original seed.
        seed_length = true_query_end - true_query_start
        if seed_length > self.max_seed_length:
            initial_true_query_end = true_query_end
            true_query_end = true_query_start + self.max_seed_length
        else:
            initial_true_query_end = true_query_end

        als = []
        for is_reverse in [True, False]:
            if is_reverse:
                seed_query_start = len(query_seq_bytes) - true_query_end
                seed_query_end = len(query_seq_bytes) - true_query_start
                possibly_RCd_query_seq = utilities.reverse_complement(query_seq_bytes)

                def covers_original_seed(query_start, query_end):
                    return query_start <= (len(query_seq_bytes) - initial_true_query_end)

            else:
                seed_query_start = true_query_start
                seed_query_end = true_query_end
                possibly_RCd_query_seq = query_seq_bytes

                def covers_original_seed(query_start, query_end):
                    return query_end >= initial_true_query_end

            seed = possibly_RCd_query_seq[seed_query_start:seed_query_end]

            for target_start in self.seed_locations[seed]:
                query_start, query_end, target_start = extend_perfect_seed(possibly_RCd_query_seq,
                                                                           self.target_seq_bytes,
                                                                           seed_query_start,
                                                                           seed_query_end,
                                                                           target_start,
                                                                           target_start + len(seed),
                                                                          )
                if not covers_original_seed(query_start, query_end):
                    continue

                cigar = [(sam.BAM_CMATCH, query_end - query_start)]
                if query_start > 0:
                    cigar = [(sam.BAM_CSOFT_CLIP, query_start)] + cigar
                if query_end < len(query_seq_bytes):
                    cigar = cigar + [(sam.BAM_CSOFT_CLIP, len(query_seq_bytes) - query_end)]

                al = pysam.AlignedSegment(self.header)
                al.cigartuples = cigar
                al.is_reverse = is_reverse
                al.is_unmapped = False
                al.reference_name = self.target_name
                al.query_sequence = possibly_RCd_query_seq
                al.query_qualities = [41] * len(query_seq_bytes)
                al.query_name = query_name

                al.reference_start = target_start

                als.append(al)

        return als

def extend_alignment(initial_al, target_seq_bytes):
    query_seq_bytes = initial_al.query_sequence.encode()
    
    if not isinstance(target_seq_bytes, bytes):
        target_seq_bytes = target_seq_bytes.encode()

    new_query_start, new_query_end, new_target_start = extend_perfect_seed(query_seq_bytes, target_seq_bytes, initial_al.query_alignment_start, initial_al.query_alignment_end, initial_al.reference_start, initial_al.reference_end)
    added_to_start = initial_al.query_alignment_start - new_query_start
    added_to_end = new_query_end - initial_al.query_alignment_end
    cigar = initial_al.cigar
    
    if added_to_start > 0:
        # Remove from starting soft clip...
        kind, length = cigar[0]
        if kind != sam.BAM_CSOFT_CLIP:
            raise ValueError(f'expected soft-clip, got {kind}')
        
        cigar[0] = (kind, length - added_to_start)

        # ... and add to subsequent match.
        kind, length = cigar[1]
        if kind != sam.BAM_CMATCH:
            raise ValueError(f'expected match, got {kind}')
            
        cigar[1] = (kind, length + added_to_start)

        if cigar[0][1] == 0:
            cigar = cigar[1:]
        
    if added_to_end > 0:
        # Remove from ending soft clip...
        kind, length = cigar[-1]
        if kind != sam.BAM_CSOFT_CLIP:
            raise ValueError(f'expected soft-clip, got {kind}')
        
        cigar[-1] = (kind, length - added_to_end)
        
        # ... and add to subsequent match.
        kind, length = cigar[-2]
        if kind != sam.BAM_CMATCH:
            raise ValueError(f'expected match, got {kind}')
            
        cigar[-2] = (kind, length + added_to_end)

        if cigar[-1][1] == 0:
            cigar = cigar[:-1]

    if added_to_start > 0 or added_to_end > 0:
        new_al = copy.deepcopy(initial_al)
        new_al.cigar = cigar
        new_al.reference_start = new_target_start
    else:
        new_al = initial_al
    
    return new_al

def extend_alignment_with_one_nt_deletion(initial_al, target_seq_bytes, extend_before=True, extend_after=True):
    query_seq_bytes = initial_al.query_sequence.encode()
    
    if not isinstance(target_seq_bytes, bytes):
        target_seq_bytes = target_seq_bytes.encode()

    gained_before, gained_after = extend_perfect_seed_with_one_nt_deletion(query_seq_bytes, target_seq_bytes, initial_al.query_alignment_start, initial_al.query_alignment_end, initial_al.reference_start, initial_al.reference_end)
    cigar = initial_al.cigar
    
    if gained_before > 0 and extend_before:
        # Remove from starting soft clip...
        kind, length = cigar[0]
        if kind != sam.BAM_CSOFT_CLIP:
            raise ValueError(f'expected soft-clip, got {kind}')
        
        soft_clip_length = length - gained_before
        if soft_clip_length > 0:
            soft_clip = [(kind, soft_clip_length)]
        else:
            soft_clip = []
        
        # ... and add match and deletion.
        new_match = (sam.BAM_CMATCH, gained_before)
        deletion = (sam.BAM_CDEL, 1)

        cigar = soft_clip + [new_match, deletion] + cigar[1:]
            
    if gained_after > 0 and extend_after:
        # Remove from ending soft clip...
        kind, length = cigar[-1]
        if kind != sam.BAM_CSOFT_CLIP:
            raise ValueError(f'expected soft-clip, got {kind}')
        
        soft_clip_length = length - gained_after
        if soft_clip_length > 0:
            soft_clip = [(kind, soft_clip_length)]
        else:
            soft_clip = []
        
        # ... and add match and deletion
        new_match = (sam.BAM_CMATCH, gained_after)
        deletion = (sam.BAM_CDEL, 1)

        cigar = cigar[:-1] + [deletion, new_match] + soft_clip
        
    if (gained_before > 0 and extend_before) or (gained_after > 0 and extend_after):
        new_al = copy.deepcopy(initial_al)
        new_al.cigar = cigar
        if (gained_before > 0 and extend_before):
            # reference_start changes by 1 for deletion plus gained_before
            new_al.reference_start = initial_al.reference_start - 1 - gained_before
    else:
        new_al = initial_al
    
    return new_al

def extend_repeatedly(initial_al, target_seq_bytes, extend_before=True, extend_after=True):
    ''' extend with 1 nt deletions until doing so doesn't change the alignment '''
    
    extended = initial_al # Confusing to call this 'extended' but helps make the while loop simpler.
    previous = None
    while extended is not previous:
        previous = extended
        extended = extend_alignment_with_one_nt_deletion(previous, target_seq_bytes, extend_before, extend_after)

    return extended



import pprint
import urllib

from hits import utilities

def parse_attribute_string(attribute_string):
    if attribute_string == '.':
        parsed = {}
    else:
        fields = attribute_string.split(';')
        pairs = [field.split('=') for field in fields]
        parsed = {name: urllib.parse.unquote(value).strip('"') for name, value in pairs}

    return parsed

def make_attribute_string(attribute):
    entries = []
    for key, value in sorted(attribute.items()):
        key = urllib.parse.quote(str(key), safe='')
        value = urllib.parse.quote(str(value), safe='')
        entry = f'{key}={value}'
        entries.append(entry)

    attribute_string = ';'.join(entries)
    return attribute_string

class Feature(object):
    def __init__(self, line=None):
        if line == None:
            # Allow __init__ to be called with no arguments to allow the
            # @classmethod constructor below.
            return

        fields = line.strip().split('\t')
        
        self.seqname = fields[0]
        self.source = fields[1]
        self.feature = fields[2]

        # Note conversion to 0-based indexing.
        self.start = int(fields[3]) - 1
        self.end = int(fields[4]) - 1

        self.score = fields[5]
        self.strand = fields[6]
        self.frame = fields[7]
        if self.frame != '.':
            self.frame = int(self.frame)
        self.attribute_string = fields[8]
        
        self.parse_attribute_string()

        self.parent = None
        self.children = set()
    
    @classmethod
    def from_fields(cls,
                    seqname='.',
                    source='.',
                    feature='.',
                    start='.',
                    end='.',
                    score='.',
                    strand='.',
                    frame='.',
                    attribute_string='.',
                    ID=None,
                   ):
        obj = cls()
        obj.seqname = seqname
        obj.source = source
        obj.feature = feature
        obj.start = start
        obj.end = end
        obj.score = score
        obj.strand = strand
        obj.frame = frame
        obj.attribute_string = attribute_string
        obj.parse_attribute_string()
        if ID is not None:
            obj.attribute['ID'] = ID
        return obj

    @classmethod
    def sequence_edge(cls, seqname, position):
        obj = cls()
        obj.seqname = seqname
        obj.feature = 'edge'
        obj.start = position
        obj.end = position
        obj.strand = '.'
        obj.source = '.'
        obj.score = '.'
        obj.frame = '.'
        obj.attribute = {'ID': f'edge_{seqname}_{position}'}
        obj.populate_attribute_string()
        return obj

    def parse_attribute_string(self):
        self.attribute = parse_attribute_string(self.attribute_string)

    def populate_attribute_string(self):
        self.attribute_string = make_attribute_string(self.attribute)

    def populate_connections(self, id_to_object):
        parent_id = self.attribute.get('Parent')
        if parent_id:
            self.parent = id_to_object[parent_id]
            self.parent.children.add(self)

    @property
    def ID(self):
        return self.attribute.get('ID')

    @property
    def descendants(self):
        descendants = set(self.children)

        for child in self.children:
            descendants.update(child.descendants)

        return descendants

    def print_family(self, level=0):
        indent = '\t' * min(level, 3)

        print(f'{indent}{self}')

        if level == 0:
            pprint.pprint(self.attribute)

        for child in sorted(self.children):
            child.print_family(level=level + 1)
    
    def __str__(self):
        self.populate_attribute_string()
        fields = (self.seqname,
                  self.source,
                  self.feature,
                  str(self.start + 1), # Note conversion back to 1-based indexing.
                  str(self.end + 1),
                  self.score,
                  self.strand,
                  str(self.frame),
                  self.attribute_string,
                 )
        line = '\t'.join(fields)
        return line
    
    @property
    def pasteable(self):
        return f'{self.seqname}:{self.start}-{self.end}'
    
    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return str(self) == str(other)

    @property
    def comparison_key(self):
        key = (self.seqname,
               self.start,
               self.end,
               self.feature,
               self.strand,
              )
        return key

    def __lt__(self, other):
        return self.comparison_key < other.comparison_key

    def is_contained_in(self, other):
        return self.seqname == other.seqname and \
               self.start >= other.start and \
               self.end <= other.end

    def __len__(self):
        return self.end - self.start + 1

    def sequence(self, reference_sequences):
        ''' Given dictionary of reference_sequences which includes an entry for 
        self.seqname, returns stranded sequence for this feature.
        '''
        seq = reference_sequences[self.seqname][self.start:self.end + 1]
        if self.strand == '-':
            seq = utilities.reverse_complement(seq)

        return seq

def populate_all_connections(features):
    for f in features:
        f.children = set()
        f.parent = None

    id_to_object = {f.attribute['ID']: f for f in features if 'ID' in f.attribute}

    for f in features:
        f.populate_connections(id_to_object)

def get_all_features(gff_fn, populate_connections=True):
    def relevant_lines(gff_fn):
        # Ignore any line starting with '#' and all lines after any line
        # starting with '##FASTA'
        with open(gff_fn) as gff_fh:
            for line in gff_fh:
                if line.startswith('##FASTA'):
                    break
                elif line.startswith('#'):
                    continue
                else:
                    yield line

    all_features = [Feature(line) for line in relevant_lines(gff_fn)]

    if populate_connections:
        populate_all_connections(all_features)

    return all_features

def get_top_level_features(features):
    populate_all_connections(features)
    top_level_features = [f for f in features if f.parent == None]
    return top_level_features



import re
import numpy as np
import glob
import os
import argparse
import subprocess
from itertools import izip, product
from struct import unpack

import tqdm

def line_count(file_name):
    ''' Returns the number of lines in file_name by calling wc. '''
    wc_output = subprocess.check_output(['wc', '-l', file_name])
    count, _ = wc_output.strip().split()
    count = int(count)
    return count

def custom_progress_bar(max_val):
    if progressbar:
        max_str = str(len(str(max_val)))
        format_string = '%(value)' + max_str + 'd / %(max)d'
        widgets = [progressbar.Bar('='),
                   ' ',
                   progressbar.FormatLabel(format_string),
                   ' ',
                   progressbar.ETA(),
                  ]
        progress_bar = progressbar.ProgressBar(widgets=widgets, maxval=max_val).start()
    else:
        # If progressbar isn't installed, create a dummy class that will look like a progressbar but do nothing
        class Dummy(object):
            def update(self, number):
                pass

            def finish(self):
                pass

        progress_bar = Dummy()

    return progress_bar

def load_locations(lane_dir):
    ''' Loads cluster location information from .locs files of the form
        s_1_{tile number}.locs in lane_dir. 
        
        Returns a dict with key-value pairs of the form 
        {tile number: location dict}, where location dict has key-value pairs
        of the form {x, y: cluster number}.
        
        .locs format description, copied from picard documentation
        bytes 0-3 : (int?) Version number (1) 
        bytes 4-7 : 4 byte float equaling 1.0
        bytes 8-11 : unsigned int numClusters
        bytes 12-15: : X coordinate of first cluster (32-bit float)
        bytes 16-19: : Y coordinate of first cluster (32-bit float)
        The remaining bytes of the file store the X and Y coordinates of the remaining clusters.
    '''
    def transform_coordinate(f):
        ''' Transform coordinate from .locs system to fastq system, as per
            http://www.biostars.org/p/51681/
        '''
        return int(round(10 * f)) + 1000
    
    locs_files = sorted(glob.glob('{0}/*.locs'.format(lane_dir)))
    
    locs = {}
    for i, locs_file in tqdm.tqdm_notebook(enumerate(locs_files), desc='Tiles', total=len(locs_files)):
        root, ext = os.path.splitext(locs_file)
        tile = int(root.split('_')[-1])
        
        locs_handle = open(locs_file, 'rb')
        
        version, = unpack('i', locs_handle.read(4))
        if version != 1:
            message = 'LOCS version number of {0}, expected 1'.format(version)
            raise ValueError(message)
        
        one, = unpack('f', locs_handle.read(4))
        if one != 1.0:
            message = 'LOCS float check of {0:.2f}, expected 1.0'.format(one)
            raise ValueError(message)
        
        num_clusters, = unpack('i', locs_handle.read(4))
        
        locs[tile] = {}

        #for c in tqdm.tqdm_notebook(xrange(num_clusters), desc=str(tile), total=num_clusters, leave=False):
        for c in xrange(num_clusters):
            eight_bytes = locs_handle.read(8)
            
            if len(eight_bytes) != 8:
                message = 'LOCS file contains less clusters than indicated in header'
                raise ValueError(message)
            
            x, y = unpack('ff', eight_bytes)
            x, y = transform_coordinate(x), transform_coordinate(y)
            locs[tile][x, y] = c

        if locs_handle.read(1):
            message = 'LOCS file contains more data than indicated in header'
            raise ValueError(message)
        
    return locs

def load_intensities(lane_dir, tile, num_clusters):
    ''' Loads intensity values from .cif files located in subdirectories of
        lane_dir. Returns a 3D array indexed by 
        (cluster number, base identity, cycle number).
    
        .cif format, copied from HCS 1.4/RTA 1.12 Theory of Operation 
        bytes 0-2: CIF
        byte 3: Version number (1)
        byte 4: Precision. Can be 1 for a file storing intensities as signed
                bytes, 2 for values stored as signed 2-byte integers, or 4 for
                values stored as 4-byte floating-point values. Normal .cif
                files use 2 bytes of precision.
        bytes 5-6: Cycle (unsigned short)
        bytes 7-8: 1 (unsigned short)
        bytes 9-12: Cluster count (unsigned int)
        The remainder of the file stores the A intensities, then C, then G,
        then T. The intensities for each channel take up 
        (Precision * ClusterCount) bytes.
    '''
    cycle_dirs = glob.glob('{0}/C*'.format(lane_dir))
    cycle_pattern = re.compile(r'C(\d+)\.1')
    extract_cycle = lambda s: int(cycle_pattern.search(s).group(1))
    cycles = sorted([extract_cycle(cycle_dir) for cycle_dir in cycle_dirs])
    
    intensities = np.zeros(shape=(num_clusters, 4, len(cycles)), dtype=np.int)
    
    progress_bar = custom_progress_bar(len(cycles))
    for cycle in cycles:
        tile_file = '{0}/C{1}.1/s_1_{2}.cif'.format(lane_dir,
                                                    cycle,
                                                    tile,
                                                   )
        with open(tile_file, 'rb') as tile_handle:
            cif = tile_handle.read(3)
            if cif != 'CIF':
                message = '\'CIF\' expected in CIF header, got {0}'.format(cif)
                raise ValueError(message)
            
            version, = unpack('b', tile_handle.read(1))
            if version != 1:
                message = 'CIF version number of {0}, expected 1'.format(version)
                raise ValueError(message)
            
            precision, = unpack('b', tile_handle.read(1))
            if precision != 2:
                message = 'Support for CIF precision of {0} not implemented'.format(precision)
                raise NotImplementedError(message)

            cycle_claimed, = unpack('H', tile_handle.read(2))
            if cycle_claimed != cycle:
                message = 'Cycle of {0} claimed in CIF header, expected {1}'.format(cycle_claimed,
                                                                                    cycle,
                                                                                   )
                raise ValueError(message)

            # Next 2 bytes don't seem to agree with the standard. Ignore them.
            one = unpack('H', tile_handle.read(2))

            num_clusters_claimed, = unpack('I', tile_handle.read(4))
            if num_clusters_claimed != num_clusters:
                message = '{0} clusters claimed in CIF header, expected {1}'.format(num_clusters_claimed,
                                                                                    num_clusters,
                                                                                   )
                raise ValueError(message)

            # Order in .cif is ACGT
            format_string = '{0}h'.format(num_clusters)
            for base in range(4):
                all_clusters = tile_handle.read(num_clusters * 2)
                if len(all_clusters) != num_clusters * 2:
                    message = 'CIF file ended prematurely'
                    raise ValueError(message)
                      
                all_intensities = unpack(format_string, all_clusters)
                for cluster, intensity in enumerate(all_intensities):
                    intensities[cluster, base, cycle - 1] = intensity
            
            # Make sure file ended when it said it would
            if tile_handle.read(1):
                message = 'CIF file contains more data than indicated in header'
                raise ValueError(message)

        progress_bar.update(cycle)
    progress_bar.finish()

    return intensities

def intensities_array_to_strings(intensities_array):
    def base_to_string(base):
        strings = map(str, base)
        string = '\t'.join(strings) + '\n'
        return string
    
    def cluster_to_string(cluster):
        base_strings = map(base_to_string, cluster)
        string = ''.join(base_strings)
        return string
    
    num_clusters, _, _ = intensities_array.shape
    intensity_strings = []
    progress_bar = custom_progress_bar(num_clusters)
    for c, cluster in enumerate(intensities_array):
        string = cluster_to_string(cluster)
        intensity_strings.append(string)
        
        progress_bar.update(c + 1)
    progress_bar.finish()

    return intensity_strings

def make_intensity_files(fastq_files, lane_dir):
    name_delimiter = re.compile(r'[ @]')
    def parse_name_line(name_line):
        name = name_line.strip().lstrip('@')
        cluster_info, read_info = re.split(name_delimiter, name)
        tile, x, y = cluster_info.split(':')[-3:]
        tile, x, y = int(tile), int(x), int(y)
        return cluster_info, tile, x, y

    def split_int_file_name(fastq_file, tile):
        root, ext = os.path.splitext(fastq_file)
        int_file = '{0}_int_{1}.txt'.format(root, tile)
        return int_file

    def merged_int_file_name(fastq_file):
        root, ext = os.path.splitext(fastq_file)
        int_file = '{0}_int.txt'.format(root, tile)
        return int_file

    fastq_line_counts = {}
    print('Counting reads in fastq files...')
    progress_bar = custom_progress_bar(len(fastq_files))
    for i, fastq_file in enumerate(fastq_files):
        fastq_line_counts[fastq_file] = line_count(fastq_file) / 4
        progress_bar.update(i + 1)
    progress_bar.finish()
    
    print('Loading cluster locations in tiles...')
    locations = load_locations(lane_dir)
    cluster_counts = {tile:len(locations) for tile, locations in locations.items()}
    
    for t, tile in enumerate(locations):
        print('Processing tile {0} ({1}/{2})'.format(tile, t + 1, len(locations)))
        print('Loading intensities from .cifs...')
        intensities = load_intensities(lane_dir, tile, cluster_counts[tile])
        print('Converting to strings...')
        intensity_strings = intensities_array_to_strings(intensities)
        
        for fastq_file in fastq_files:
            print('Assigning intensities to {0}'.format(fastq_file))
            
            intensities_file_name = split_int_file_name(fastq_file, tile)

            fastq_handle = open(fastq_file)
            fastq_reads = izip(*[fastq_handle]*4)
            name_lines = (name_line for name_line, _, _, _ in fastq_reads)
            
            with open(intensities_file_name, 'w') as intensities_handle:
                progress_bar = custom_progress_bar(fastq_line_counts[fastq_file])
                for l, name_line in enumerate(name_lines):
                    cluster_info, read_tile, x, y = parse_name_line(name_line)
                    
                    if read_tile == tile:
                        if (x, y) not in locations[tile]:
                            # Attempt to deal with bizarre rounding irregularities
                            valid_nearby = []
                            offsets = [-1, 0, 1]
                            for x_offset, y_offset in product(offsets, offsets):
                                if (x + x_offset, y + y_offset) in locations[tile]:
                                    valid_nearby.append((x + x_offset, y + y_offset))
                            
                            if len(valid_nearby) != 1:
                                raise ValueError('Unable to resolve rounding irregularity')
                            
                            x, y = valid_nearby[0]
                        
                        cluster_index = locations[tile][x, y]
                        intensity_string = intensity_strings[cluster_index]
                        
                        intensities_handle.write(name_line)
                        intensities_handle.write(intensity_string)
                    
                    progress_bar.update(l + 1)
                progress_bar.finish()
                 
    for fastq_file in fastq_files:
        print('Merging tiles for {0}'.format(fastq_file))
        tile_handles = {tile: open(split_int_file_name(fastq_file, tile))
                        for tile in locations}
        
        merged_file = merged_int_file_name(fastq_file)
        
        fastq_handle = open(fastq_file)
        fastq_reads = izip(*[fastq_handle]*4)
        name_lines = (name_line for name_line, _, _, _ in fastq_reads)
        
        with open(merged_file, 'w') as merged_handle:
            progress_bar = custom_progress_bar(fastq_line_counts[fastq_file])
            
            for l, name_line in enumerate(name_lines):
                cluster_info, read_tile, x, y = parse_name_line(name_line)
                tile_handle = tile_handles[read_tile]
                
                tile_name_line = tile_handle.readline()
                if tile_name_line != name_line:
                    raise RuntimeError('Split files out of sync')

                merged_handle.write('@{0}\n'.format(cluster_info))
                for i in range(4):
                    merged_handle.write(tile_handle.readline())
                
                progress_bar.update(l + 1)
            progress_bar.finish()
        
        print('Removing tile-specific files...')
        for tile in locations:
            os.remove(split_int_file_name(fastq_file, tile))

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('fastq_files',
                        help='comma-separated list of fastq file names to extract the intensities of the reads in',
                        type = lambda s: s.split(','),
                       )
    parser.add_argument('lane_dir',
                        help='directory containing .loc files for each tile and subdirectories for each cycle containing .cif files',
                       )
    args = parser.parse_args()

    make_intensity_files(args.fastq_files, args.lane_dir)



from itertools import cycle

import numpy as np
import pandas as pd
import bokeh.plotting
import bokeh.io
import bokeh.models

from .external_coffeescript import build_callback

color_list = bokeh.palettes.Category20[20]

def pc(df):
    fig = bokeh.plotting.figure(width=1500,
                                height=1200,
                                tools=['reset', 'undo', 'save', 'pan', 'wheel_zoom', 'tap'],
                               )

    box_select = bokeh.models.tools.BoxSelectTool(name='box_select')
    box_select.callback = build_callback('pc_box_select')
    fig.add_tools(box_select)
    fig.toolbar.active_drag = box_select

    if df.columns.nlevels == 2:
        names = ['{0}: {1}'.format(gn, en) for gn, en in df.columns.values]
    else:
        names = df.columns

    lines = []
    for y, ((row_name, row), color) in enumerate(zip(df.iterrows(), cycle(color_list))):
        vals = row.values
        data = {
            'x': list(range(len(vals))),
            'y': vals,
            'name': [row_name]*len(names),
            'dq': [False for _ in vals],
        }

        source = bokeh.models.ColumnDataSource(data=data, name='source_by_y_{0}'.format(y))

        line = fig.line(x='x', y='y',
                        source=source,
                        line_alpha=0.7,
                        line_width=3,
                        color=color,
                        name='line_{0}'.format(y),
                       )
        line.hover_glyph = line.glyph
        lines.append(line)

    columns = df.columns.values

    colors = [c for c, _ in zip(cycle(color_list), df.index)]

    for x, exp_name in enumerate(columns):
        vals = df[exp_name].values
        data = {
            'x': [x for _ in vals],
            'y': vals,
            'dq': [False for _ in vals],
            'color': colors,
            'size': [6 for _ in vals],
            'alpha': [0.9 for _ in vals],
        }

        source = bokeh.models.ColumnDataSource(data=data, name='source_by_x_{0}'.format(x))
        circles = fig.circle(x='x', y='y',
                             source=source,
                             color='color',
                             fill_alpha='alpha',
                             line_alpha=0,
                             size='size',
                            )
        source.callback = build_callback('pc_selection')
        
    for x in [0, len(columns) - 1]:
        vals = df.iloc[:, x].values

        if x == 0:
            text_align = 'right'
            x_offset = -10
        else:
            text_align = 'left'
            x_offset = 10
            
        data = {
            'x': [x for _ in vals],
            'y': vals,
            'label': df.index.values,
            'hover_color': colors,
            'text_color': ['black' for _ in vals],
            'text_alpha': [0.9 for _ in vals],
            'highlight': [False for _ in vals],
            'dq': [False for _ in vals],
        }

        source = bokeh.models.ColumnDataSource(data=data)

        labels = bokeh.models.LabelSet(x='x',
                                       y='y',
                                       text='label',
                                       level='glyph',
                                       x_offset=x_offset,
                                       y_offset=0,
                                       source=source,
                                       text_font_size='12pt',
                                       text_align=text_align,
                                       text_alpha='text_alpha',
                                       text_color='text_color',
                                       text_baseline='middle',
                                       name='labels',
                                       text_font='monospace',
                                      )
        fig.add_layout(labels)
        
    hover = bokeh.models.HoverTool(line_policy='interp',
                                   renderers=lines,
                                   tooltips=None,
                                   name='hover_tool',
                                  )
    hover.callback = build_callback('pc_hover')
    fig.add_tools(hover)
    fig.xgrid.visible = False

    fig.xaxis.minor_tick_line_color = None

    code = '''
    dict = {dict}
    return dict[tick]
    '''.format(dict=dict(enumerate(names)))

    fig.xaxis.ticker = bokeh.models.FixedTicker(ticks=list(range(len(names))))
    fig.xaxis.formatter = bokeh.models.FuncTickFormatter(code=code)
    fig.xaxis.major_label_text_font_size = '12pt'
    fig.xaxis.major_label_orientation = np.pi / 4

    fig.yaxis.axis_label = 'nts between cut'
    fig.yaxis.axis_label_text_font_style = 'normal'
    fig.yaxis.axis_label_text_font_size = '16pt'
    fig.yaxis.major_label_text_font_size = '12pt'

    fig.x_range = bokeh.models.Range1d(-0.1 * len(names), 1.1 * (len(names) - 1))
    fig.x_range.callback = bokeh.models.CustomJS.from_coffeescript(code=f'''
    if cb_obj.start < {-0.1 * len(names)}
        cb_obj.start = {-0.1 * len(names)}

    if cb_obj.end > {1.1 * (len(names) - 1)}
        cb_obj.end = {1.1 * (len(names) - 1)}
    ''')

    fig.y_range.callback = bokeh.models.CustomJS.from_coffeescript(code=f'''
    if cb_obj.start < 0
        cb_obj.start = 0
        
    if cb_obj.end > {df.max().max() * 1.05}
        cb_obj.end = {df.max().max() * 1.05} 
    ''')

    clear_constraints = bokeh.models.widgets.Button(label='Clear constraints',
                                                    width=50,
                                                    name='clear_constraints',
                                                   )
    clear_constraints.callback = build_callback('pc_clear')

    data = dict(x=[], width=[], top=[], bottom=[])
    constraint_source = bokeh.models.ColumnDataSource(data=data, name='constraints')
    vbar = bokeh.models.glyphs.VBar(x='x',
                                    top='top',
                                    bottom='bottom',
                                    width='width',
                                    fill_alpha=0.15,
                                    fill_color='black',
                                    line_width=0,
                                    line_alpha=0,
                                   )

    fig.add_glyph(constraint_source, vbar, selection_glyph=vbar, nonselection_glyph=vbar)

    fig.min_border = 100

    text_input = bokeh.models.widgets.TextInput(title='Highlight:', name='search')
    text_input.callback = build_callback('pc_search')

    columns = [
        bokeh.layouts.column([fig]),
        bokeh.layouts.column([bokeh.layouts.Spacer(height=100), clear_constraints, text_input]),
    ]

    bokeh.io.show(bokeh.layouts.row(columns))



import numpy as np
import pysam
import string
from itertools import chain, islice, zip_longest

from . import utilities
from . import fastq
from . import fasta
from . import genomes
from . import mapping_tools
from . import sw
from . import sam

def mapping_to_alignment(mapping, base_lookup):
    ''' Convert a mapping represented by a pysam.AlignedRead into an alignment. '''
    path = []
    mismatches = set()
    deletions = set()

    for read_i, ref_i in sam.aligned_pairs_exclude_soft_clipping(mapping):
        if read_i != None:
            if ref_i == None:
                ref_i = sw.GAP
            else:
                read_base = mapping.seq[read_i]
                ref_base = base_lookup(mapping.reference_name, ref_i).upper()
                if read_base != ref_base:
                    mismatches.add((read_i, ref_i))

            path.append((read_i, ref_i))
        else:
            deletions.add(ref_i)
    
    alignment = {
        'path': path,
        'is_reverse': mapping.is_reverse,
        'mismatches': mismatches,
        'deletions': deletions,
        'rname': mapping.reference_name,
        'query': mapping.seq,
    }

    for tag in ['XO', 'XM']:
        if mapping.has_tag(tag):
            alignment[tag] = mapping.get_tag(tag)
        else:
            alignment[tag] = -1

    return alignment

def produce_bowtie2_alignments(reads,
                               index_prefix,
                               genome_dir,
                               score_min,
                               max_alignments=5,
                              ):

    bowtie2_options = {
        'local': True,
        'report_up_to': 10,
        'seed_mismatches': 1,
        'seed_interval_function': 'C,1,0',
        'seed_length': 10,
    }

    sam_file, mappings = mapping_tools.map_bowtie2(index_prefix,
                                                   reads=reads,
                                                   custom_binary=True,
                                                   score_min=score_min,
                                                   yield_mappings=True,
                                                   **bowtie2_options)

    base_lookup = genomes.build_base_lookup(genome_dir)
    mapping_groups = utilities.group_by(mappings, lambda m: m.qname)
    
    for qname, group in mapping_groups:
        mapped = [m for m in group if not m.is_unmapped]
        mapped = sorted(mapped, key=lambda m: m.get_tag('AS'), reverse=True)[:max_alignments]
        group = sorted(mapped, key=lambda m: (m.reference_name, m.pos))
        alignments = [mapping_to_alignment(mapping, base_lookup)
                      for mapping in mapped]
        yield qname, alignments

def get_local_alignments(read, targets):
    seq = read.seq
    seq_rc = utilities.reverse_complement(read.seq)
    all_alignments = []
    for target in targets:
        min_score = min(20, 2 * len(target.seq))
        for query, is_reverse in [(seq, False), (seq_rc, True)]:
            alignments = sw.generate_alignments(query,
                                                target.seq,
                                                'local',
                                                min_score=min_score,
                                                max_alignments=3,
                                               )
            #if '10872' in read.name:
            #    print(targets)
            #    print(alignments)
            #    print()
            for alignment in alignments:
                if alignment['score'] >= 0.75 * 2 * len(alignment['path']):
                    alignment['query'] = query
                    alignment['rname'] = target.name
                    alignment['is_reverse'] = is_reverse
                    all_alignments.append(alignment)

    return all_alignments

def get_edge_alignments(read, targets):
    seq = read.seq
    seq_rc = utilities.reverse_complement(read.seq)
    all_alignments = []
    min_score = 10
    for target in targets:
        for query, is_reverse in [(seq, False), (seq_rc, True)]:
            alignments = sw.generate_alignments(query,
                                                target.seq,
                                                'IVT',
                                                min_score=min_score,
                                                max_alignments=1,
                                                mismatch_penalty=-2,
                                               )
            for alignment in alignments:
                if alignment['score'] >= 2 * len(alignment['path']):
                    alignment['query'] = query
                    alignment['rname'] = target.name
                    alignment['is_reverse'] = is_reverse
                    all_alignments.append(alignment)

    return all_alignments

def up_to_first_space(string):
    beginning = string.split(' ')[0]
    return beginning

def produce_sw_alignments(reads, genome_dirs, extra_targets, max_to_report=5):
    targets = set()

    for genome_dir in genome_dirs:
        fasta_fns = genomes.get_all_fasta_file_names(genome_dir)
        for fasta_fn in fasta_fns:
            targets.update(fasta.reads(fasta_fn))

    targets.update(extra_targets)

    for read in reads:
        alignments = get_local_alignments(read, targets) + get_edge_alignments(read, targets)
        # bowtie2 only retains up to the first space in a qname, so do the same
        # here to allow qnames to be compared
        alignments = sorted(alignments, key=lambda a: a['score'], reverse=True)
        alignments = alignments[:max_to_report]

        name = read.name

        sanitized_name = up_to_first_space(name)
        yield sanitized_name, alignments

def produce_sam_alignments(bam_fn, ref_fn, max_reads=None):
    refs = {r.name: r.seq for r in fasta.reads(ref_fn)}
    def base_lookup(name, i):
        return refs[name][i:i + 1]
    
    mapping_groups = utilities.group_by(pysam.AlignmentFile(str(bam_fn)), lambda m: m.query_name)
    for name, mappings in islice(mapping_groups, max_reads):
        alignments = [mapping_to_alignment(mapping, base_lookup)
                      for mapping in sorted(mappings, key=lambda m: (m.tid, m.pos))
                      if not mapping.is_unmapped
                     ]
        yield name, alignments

def produce_representations(alignment_groups_list):
    for alignment_group_list in zip_longest(*alignment_groups_list):
        representations = []
        qnames = set()
        for qname, alignment_group in alignment_group_list:
            qnames.add(qname)
            representations.extend(map(represent_alignment, alignment_group))

        if len(qnames) > 1:
            print([(qname, len(group)) for qname, group in alignment_group_list])
            raise ValueError('Attempted to flatten alignment groups with different qnames:', qnames)

        qname = qnames.pop()

        yield qname, representations

def find_best_offset(R1, R2_rc):
    R2 = R2_rc.reverse_complement()
    status, insert_length, alignment = sw.infer_insert_length(R1, R2, '', '')
    if status == 'good':
        offset = insert_length - len(R1)
    else:
        offset = len(R1) + 1
    return offset

def represent_alignment(alignment):
    ''' Returns text representation of the portion of the read mapped by a local
    mapping and the genomic region mapped to. Also returns a key for sorting
    such representations by how far from the left edge they begin.
    '''
    read_positions = {}
    read_to_ref = {}
    query_length = len(alignment['query'])
    for read_i, ref_i in alignment['path']:
        if read_i != sw.GAP:
            if alignment['is_reverse']:
                absolute_read_position = query_length - 1 - read_i
                match = '<'
            else:
                absolute_read_position = read_i
                match = '>'
            if ref_i != sw.GAP:
                if (read_i, ref_i) in alignment['mismatches']:
                    match = 'x'

                if ref_i - 1 in alignment['deletions']:
                    match = '\\'
                elif ref_i + 1 in alignment['deletions']:
                    match = '/'

                read_positions[absolute_read_position] = match
                read_to_ref[absolute_read_position] = ref_i
            else:
                read_positions[absolute_read_position] = '-'

    leftmost_read = min(read_positions)
    rightmost_read = max(read_positions)
    read_positions[leftmost_read] = '|'
    read_positions[rightmost_read] = '|'
    
    before_left = ' '*leftmost_read
    after_right = ' '*(query_length - rightmost_read - 1)

    width = rightmost_read - leftmost_read + 1

    ref_name = alignment['rname']

    ref_name_string = ref_name.center(width)
    if width > 100:
        parts = [' ',
                 ref_name,
                 ref_name_string[1 + len(ref_name):-(len(ref_name) + 1) ],
                 ref_name,
                 ' ',
                ]
        ref_name_string = ''.join(parts)

    ref_name_line = before_left + ref_name_string + after_right
    
    # Create a line marking the reference positions of the boundaries of the
    # mapping. If there is room, summarize the mismatches and indel of the
    # mapping in between.
    if alignment['XM'] > 0 or alignment['XO'] > 0:
        mismatch_string = ' XM={0}, XO={1} '.format(alignment['XM'],
                                                    alignment['XO'],
                                                   )
    else:
        mismatch_string = ''

    mismatch_chars = list(mismatch_string.center(width, '-'))

    right_ref_edge_string = '{0:,}'.format(read_to_ref[rightmost_read])
    left_ref_edge_string = '{0:,}'.format(read_to_ref[leftmost_read])

    left_edge_slice = slice(1, len(left_ref_edge_string) + 1)
    right_edge_slice = slice(-(len(right_ref_edge_string) + 1), -1)

    overlaps_left = any(c != '-' for c in mismatch_chars[left_edge_slice])
    overlaps_right = any(c != '-' for c in mismatch_chars[right_edge_slice])

    if overlaps_left or overlaps_right:
        # If the verbose version overlaps, try to make a smaller string.
        mismatch_string = ' {0} {1} '.format(alignment['XM'],
                                             alignment['XO'],
                                            )
        mismatch_chars = list(mismatch_string.center(width, '-'))

        overlaps_left = any(c != '-' for c in mismatch_chars[left_edge_slice])
        overlaps_right = any(c != '-' for c in mismatch_chars[right_edge_slice])

        if overlaps_left or overlaps_right:
            mismatch_string = ''
    
    mismatch_chars = list(mismatch_string.center(width))
    extent_chars = mismatch_chars

    left_start, left_stop, _ = left_edge_slice.indices(len(extent_chars))
    right_start, right_stop, _ = right_edge_slice.indices(len(extent_chars))
    if left_stop < right_start:
        extent_chars[left_edge_slice] = list(left_ref_edge_string)
        extent_chars[right_edge_slice] = list(right_ref_edge_string)
    extent_line = before_left + ''.join(extent_chars) + after_right
    
    read_positions_chars = [' ']*query_length
    for p in read_positions:
        read_positions_chars[p] = read_positions[p]
    read_positions_line = ''.join(read_positions_chars)

    lines = (ref_name_line, extent_line, read_positions_line)
    
    return frozenset(read_positions), lines

def combine_representations(first_read_positions,
                            first_lines,
                            second_read_positions,
                            second_lines,
                           ):
    ''' Combiner to be used during the collapsing process. '''
    combined_read_positions = first_read_positions | second_read_positions
    arrays = [[' ' for p in range(len(first_lines[0]))] for l in range(len(first_lines))]
    for p in first_read_positions:
        for l, line in enumerate(first_lines):
            arrays[l][p] = line[p]
    for p in second_read_positions:
        for l, line in enumerate(second_lines):
            arrays[l][p] = line[p]
    combined_lines = [''.join(array) for array in arrays]
    return combined_read_positions, combined_lines

def leftmost_position(representation):
    read_positions, lines = representation
    return min(read_positions)

def positions_covered(representation):
    read_positions, lines = representation
    return len(read_positions)

def collapse_representations(representations):
    ''' Push nonoverlapping representation down into the same line. '''
    if not representations:
        return representations

    # Sort by leftmost position prior to collapsing
    representations = sorted(set(representations), key=leftmost_position)
    
    collapsed = [representations[0]]
    for read_positions, lines in representations[1:]:
        was_collapsed = False
        for t, (target_positions, target_lines) in enumerate(collapsed):
            if read_positions & target_positions == set():
                collapsed[t] = combine_representations(read_positions, lines, target_positions, target_lines)
                was_collapsed = True
                break

        if not was_collapsed:
            collapsed.append((read_positions, lines))
    
    # Sort by total positions covered after collapsing
    collapsed = sorted(collapsed, key=positions_covered)

    return collapsed

def lowercase_below_qual_threshold(seq, qual, threshold):
    ''' Returns seq with characters made lowercase at any position for which
    qual is below threshold.
    '''
    seq = list(seq)
    decoded_qual = fastq.decode_sanger(qual)

    for p, (s, q) in enumerate(zip(seq, decoded_qual)):
        if q <= threshold:
            seq[p] = s.lower()

    return ''.join(seq)

def visualize_unpaired_alignments(get_reads,
                                  sw_genome_dirs,
                                  extra_targets,
                                  bowtie2_targets,
                                  output_fn=None,
                                  skip_initial=0,
                                  num_pairs=100,
                                 ):
    def relevant_reads(source):
        return islice(source, skip_initial, skip_initial + num_pairs)

    if isinstance(get_reads, str):
        R1_fn = get_reads
        def get_reads():
            return relevant_reads(fastq.reads(R1_fn))
    else:
        full_get_reads = get_reads
        def get_reads():
            return relevant_reads(full_get_reads())

    R1_alignment_groups_list = []

    for genome_dir, index_prefix, score_min in bowtie2_targets:
        R1_alignment_groups = produce_bowtie2_alignments(get_reads(),
                                                         index_prefix,
                                                         genome_dir,
                                                         score_min,
                                                        )
        R1_alignment_groups_list.append(R1_alignment_groups)
        
    R1_sw_alignment_groups = produce_sw_alignments(get_reads(),
                                                   sw_genome_dirs,
                                                   extra_targets,
                                                  )
    R1_alignment_groups_list.append(R1_sw_alignment_groups)
    
    R1_representation_groups = produce_representations(R1_alignment_groups_list)

    everything = [get_reads(),
                  R1_representation_groups,
                 ]

    with utilities.possibly_fn(output_fn) as output_fh:
        for R1, (R1_qname, R1_representations) in zip_longest(*everything):
            if up_to_first_space(R1.name) != R1_qname:
                raise ValueError('Iters out of sync', R1.name, R1_qname)

            R1_seq = lowercase_below_qual_threshold(R1.seq, R1.qual, 20)
            
            R1_representations = collapse_representations(R1_representations)

            output_fh.write(R1.name + '\n\n')

            for _, lines in R1_representations:
                for line in lines:
                    output_fh.write(line + '\n')

            output_fh.write(R1_seq + '\n')
            output_fh.write('\n\n')

def visualize_bam_alignments(bam_fn, ref_fn, output_fn, max_reads=None):
    def get_reads():
        all_reads = sam.sam_to_fastq(bam_fn)
        grouped = utilities.group_by(all_reads, lambda r: r.name)
        for name, group in islice(grouped, max_reads):
            yield group[0]

    alignment_groups_list = [produce_sam_alignments(bam_fn, ref_fn, max_reads)]
    representation_groups = produce_representations(alignment_groups_list)

    everything = [
        get_reads(),
        representation_groups,
    ]

    with utilities.possibly_fn(output_fn) as output_fh:
        for read, (qname, representations) in zip_longest(*everything):
            if up_to_first_space(read.name) != qname:
                raise ValueError('Iters out of sync', read.name, qname)

            seq = lowercase_below_qual_threshold(read.seq, read.qual, 20)
            
            representations = collapse_representations(representations)

            output_fh.write(read.name + '\n\n')

            for _, lines in representations:
                for line in lines:
                    output_fh.write(line + '\n')

            output_fh.write(seq + '\n')
            output_fh.write('\n\n')

def visualize_paired_end_mappings(get_read_pairs,
                                  sw_genome_dirs,
                                  extra_targets,
                                  bowtie2_targets,
                                  output_fn=None,
                                  skip_initial=0,
                                  num_pairs=100,
                                 ):

    R1_alignment_groups_list = []
    R2_alignment_groups_list = []

    def relevant_reads(source):
        return islice(source, skip_initial, skip_initial + num_pairs)

    if isinstance(get_read_pairs, tuple):
        R1_fn, R2_fn = get_read_pairs
        def get_R1_reads():
            return relevant_reads(fastq.reads(R1_fn))
        
        def get_R2_rc_reads():
            return relevant_reads(fastq.reverse_complement_reads(R2_fn))
    else:
        def get_R1_reads():
            read_pairs = relevant_reads(get_read_pairs())
            return (R1 for R1, R2 in read_pairs)
        
        def get_R2_rc_reads():
            read_pairs = relevant_reads(get_read_pairs())
            return (R2.reverse_complement() for R1, R2 in read_pairs)

    for genome_dir, index_prefix, score_min in bowtie2_targets:
        R1_alignment_groups = produce_bowtie2_alignments(get_R1_reads(),
                                                         index_prefix,
                                                         genome_dir,
                                                         score_min,
                                                        )
        R1_alignment_groups_list.append(R1_alignment_groups)
        
        # Design decisions made in the parsing make it easier if R2 reads are
        # reverse complemented before mapping.
        R2_alignment_groups = produce_bowtie2_alignments(get_R2_rc_reads(),
                                                         index_prefix,
                                                         genome_dir,
                                                         score_min,
                                                        )
        R2_alignment_groups_list.append(R2_alignment_groups)

    R1_sw_alignment_groups = produce_sw_alignments(get_R1_reads(),
                                                   sw_genome_dirs,
                                                   extra_targets,
                                                  )
    R1_alignment_groups_list.append(R1_sw_alignment_groups)
    
    # Design decisions made in the parsing make it easier if R2 reads are
    # reverse complemented before mapping.
    R2_sw_alignment_groups = produce_sw_alignments(get_R2_rc_reads(),
                                                   sw_genome_dirs,
                                                   extra_targets,
                                                  )
    R2_alignment_groups_list.append(R2_sw_alignment_groups)


    R1_representation_groups = produce_representations(R1_alignment_groups_list)
    R2_representation_groups = produce_representations(R2_alignment_groups_list)

    R1_reads = get_R1_reads()
    R2_rc_reads = get_R2_rc_reads()

    everything = [R1_reads,
                  R2_rc_reads,
                  R1_representation_groups,
                  R2_representation_groups,
                 ]
    
    with utilities.possibly_fn(output_fn) as output_fh:
        for R1, R2_rc, (R1_qname, R1_representations), (R2_qname, R2_representations) in zip_longest(*everything):
            if up_to_first_space(R1.name) != R1_qname:
                raise ValueError('Iters out of sync', R1.name, R1_qname)
            if up_to_first_space(R2_rc.name) != R2_qname:
                raise ValueError('Iters out of sync', R2_rc.name, R2_qname)

            offset = find_best_offset(R1, R2_rc)

            gap_line = ''
            if offset >= 0:
                R1_shift = ''
                R2_shift = ' '*offset
            else:
                R1_shift = ' '*(-offset)
                R2_shift = ''
            
            R1_seq = lowercase_below_qual_threshold(R1.seq, R1.qual, 20)
            R2_rc_seq = lowercase_below_qual_threshold(R2_rc.seq, R2_rc.qual, 20)
            
            R1_representations = collapse_representations(R1_representations)
            R2_representations = collapse_representations(R2_representations)

            output_fh.write(R1.name + '\n')
            if R2_rc.name != R1.name:
                output_fh.write(R2_rc.name + '\n')

            for _, lines in R1_representations:
                for line in lines:
                    output_fh.write(R1_shift + line + '\n')

            output_fh.write(R1_shift + R1_seq + '\n')
            output_fh.write(gap_line)
            output_fh.write(R2_shift + R2_rc_seq + '\n')

            for _, lines in R2_representations[::-1]:
                for line in lines[::-1]:
                    output_fh.write(R2_shift + line + '\n')

            output_fh.write('\n\n')

if __name__ == '__main__1':
    bowtie2_targets = [#('/home/jah/genomes/RF_oligos', '/home/jah/bowtie2/RF_oligos', 'C,20,0'),
                       #('/home/jah/genomes/saccharomyces_cerevisiae', '/home/jah/bowtie2/saccharomyces_cerevisiae', 'C,20,0'),
                      ]
    sw_genome_dirs = ['/home/jah/genomes/truseq',
                      '/home/jah/genomes/circle_hairpins',
                      #'/home/jah/genomes/RF_oligos',
                     ]
    extra_targets = [fasta.Read('NFBC12', 'GGCTAC'),
                     fasta.Read('R', 'TGATCTCAGATCGAAAGAAGCATGGTTGTTGTTTCTGTTAGTGTAAGCAAGCGGTTTGAAAAAGAGCGCCATGAATGACTTC'),
                    ]
    
    R1_fn = '/home/jah/projects/mutations/experiments/miseq_UT_2014_11_07/data/small_R1.fastq'
    R2_fn = '/home/jah/projects/mutations/experiments/miseq_UT_2014_11_07/data/small_R2.fastq'
    results_dir = '/home/jah/projects/mutations/experiments/miseq_UT_2014_11_07/results'
    output_fn = '/home/jah/projects/mutations/experiments/miseq_UT_2014_11_07/results/visualized_mappings.txt'
  
    with open(output_fn, 'w') as output_fh:
        visualize_local_mappings(R1_fn,
                                 R2_fn,
                                 sw_genome_dirs,
                                 extra_targets,
                                 bowtie2_targets,
                                 results_dir,
                                )

if __name__ == '__main__1':
    bowtie2_targets = [('/home/jah/genomes/pCEP4_plus_e_coli', '/home/jah/bowtie2/pCEP4_plus_e_coli', 'C,20,0'),
                      ]
    sw_genome_dirs = ['/home/jah/genomes/truseq',
                      '/home/jah/genomes/circle_hairpins',
                     ]
    extra_targets = [fasta.Read('NFBC13', 'AGTCAA'),
                    ]
    
    R1_fn = '/home/jah/projects/mutations/experiments/miseq_UT_2014_09_02/CH/data/R1_contains.fastq'
    R2_fn = '/home/jah/projects/mutations/experiments/miseq_UT_2014_09_02/CH/data/R2_contains.fastq'
    output_fn = '/home/jah/projects/mutations/experiments/miseq_UT_2014_09_02/CH/results/new_visualized_mappings.txt'
  
    with open(output_fn, 'w') as output_fh:
        visualize_local_mappings(R1_fn,
                                 R2_fn,
                                 sw_genome_dirs,
                                 extra_targets,
                                 bowtie2_targets,
                                 output_fn,
                                )

if __name__ == '__main__1':
    bowtie2_targets = [('/home/jah/genomes/saccharomyces_cerevisiae', '/home/jah/bowtie2/saccharomyces_cerevisiae', 'C,20,0'),
                      ]
    sw_genome_dirs = ['/home/jah/genomes/truseq',
                      #'/home/jah/projects/ribosomes/data/guydosh_markers/',
                     ]
    extra_targets = [fasta.Read('smRNA_linker', 'CTGTAGGCACCATCAAT'),
                    ]
    
    R1_fn = '/home/jah/projects/ribosomes/experiments/guydosh_cell/wild-type_CHX/data/SRR1042853.fastq'
    
    output_fn = '/home/jah/projects/ribosomes/experiments/guydosh_cell/wild-type_CHX/results/wild-type_CHX_structures.txt'

    def get_reads():
        return islice(fastq.reads(R1_fn), 1000)


    visualize_unpaired_alignments(get_reads,
                                  sw_genome_dirs,
                                  extra_targets,
                                  bowtie2_targets,
                                  output_fn,
                                 )

if __name__ == '__main__1':
    bowtie2_targets = [('/home/jah/genomes/pCEP4_plus_contaminants/', '/home/jah/bowtie2/pCEP4_plus_contaminants', 'C,20,0'),
                       ('/home/jah/genomes/hg19/', '/home/jah/bowtie2/hg19', 'C,50,0'),
                      ]
    sw_genome_dirs = ['/home/jah/genomes/truseq',
                     ]
    extra_targets = [fasta.Read('UTBC52', 'AACATA'),
                    ]
    
    R1_unmapped_fn = '/home/jah/scratch/projects/lnt/experiments/hiseq_UT_2015_02_03/S2/results.12.00/S2_R1_unmapped.fastq'
    R2_unmapped_fn = '/home/jah/scratch/projects/lnt/experiments/hiseq_UT_2015_02_03/S2/results.12.00/S2_R2_unmapped.fastq'
    #R1_unmapped_fn = '/home/jah/scratch/projects/lnt/experiments/hiseq_UT_2015_02_03/S2/results.12.00/test_R1.fastq'
    #R2_unmapped_fn = '/home/jah/scratch/projects/lnt/experiments/hiseq_UT_2015_02_03/S2/results.12.00/test_R2.fastq'
    #R1_unmapped_fn = '/home/jah/projects/lnt/experiments/hiseq_UT_2015_02_03/S2/data/small_R1.fastq'
    #R2_unmapped_fn = '/home/jah/projects/lnt/experiments/hiseq_UT_2015_02_03/S2/data/small_R2.fastq'
    output_fn = '/home/jah/scratch/projects/lnt/experiments/hiseq_UT_2015_02_03/S2/results.12.00/'

    visualize_paired_end_mappings(R1_unmapped_fn,
                                  R2_unmapped_fn,
                                  sw_genome_dirs,
                                  extra_targets,
                                  bowtie2_targets,
                                  output_fn,
                                 )

if __name__ == '__main__':
    bowtie2_targets = [('/home/jah/genomes/saccharomyces_cerevisiae', '/home/jah/bowtie2/saccharomyces_cerevisiae', 'C,20,0'),
                      ]
    sw_genome_dirs = ['/home/jah/genomes/truseq',
                      #'/home/jah/projects/ribosomes/data/guydosh_markers/',
                     ]
    extra_targets = [fasta.Read('smRNA_linker', 'CTGTAGGCACCATCAAT'),
                    ]
    
    R1_fn = '/home/jah/projects/ribosomes/experiments/pelechano_cell/by_chx_1_r/data/SRR1919065.fastq'
    R2_fn = '/home/jah/projects/ribosomes/experiments/pelechano_cell/by_chx_1_p/data/SRR1646644_2.fastq'
    
    output_fn = '/home/jah/projects/ribosomes/experiments/pelechano_cell/by_chx_1_p/results/test.txt'

    visualize_unpaired_alignments(R1_fn,
                                  sw_genome_dirs,
                                  extra_targets,
                                  bowtie2_targets,
                                  output_fn,
                                 )
